from __future__ import annotations

import sys
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from typing import Any

from datetype import DateTime
from fritter.drivers.datetime import guessLocalZone
from fritter.drivers.sleep import SleepDriver
from fritter.persistent.json import (
    JSONableInstance,
    JSONableScheduler,
    JSONObject,
    JSONRegistry,
    LoadProcess,
    schedulerAtPath,
)

# start-registry
registry = JSONRegistry[object]()
# end-registry


# reminder-class
@dataclass
class Reminder:
    text: str

    # reminder-methods
    @classmethod
    def typeCodeForJSON(cls) -> str:
        return "reminder"

    def toJSON(
        self,
        registry: JSONRegistry[object],
    ) -> dict[str, object]:
        return {"text": self.text}

    @classmethod
    def fromJSON(
        cls,
        load: LoadProcess[object],
        json: JSONObject,
    ) -> Reminder:
        return cls(json["text"])
        # end-reminder-methods

    # app-method
    @registry.method
    def show(self) -> None:
        print(f"Reminder! {self.text}")
        # end-reminder


r: JSONableInstance[object] = Reminder("hi")


def remind(
    scheduler: JSONableScheduler[dict[Any, Any]],
    seconds: int,
    message: str,
) -> None:
    now = DateTime.now(guessLocalZone())
    later = now + timedelta(seconds=seconds)
    work = Reminder(message).show
    scheduler.callAt(later, work)


def runScheduler(newReminder: tuple[int, str] | None) -> None:
    context: dict[Any, Any] = {}
    with schedulerAtPath(
        registry,
        driver := SleepDriver(),
        Path("saved-schedule.json"),
        context,
    ) as sched:
        driver.block(1.0)
        if newReminder:
            newTime, message = newReminder
            remind(sched, newTime, message)


if __name__ == "__main__":
    args = sys.argv[1:]
    reminder = None if not args else (int(args[0]), " ".join(args[1:]))
    runScheduler(reminder)



"""
Implementation of L{PriorityQueue} in terms of the standard library's
L{heappop} and L{heappush} functions.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from heapq import heappop, heappush
from typing import Generic, Iterator, List, Optional

from .boundaries import Prioritized, PriorityQueue


@dataclass
class Heap(Generic[Prioritized]):
    """
    A simple implementation of a priority queue using the standard library's
    L{heappop} and L{heappush} functions.
    """

    _values: List[Prioritized] = field(default_factory=list)

    def add(self, item: Prioritized) -> None:
        "Implementation of L{PriorityQueue.add}"
        heappush(self._values, item)

    def get(self) -> Optional[Prioritized]:
        "Implementation of  L{PriorityQueue.get}"
        if not self._values:
            return None
        return heappop(self._values)

    def peek(self) -> Optional[Prioritized]:
        "Implementation of L{PriorityQueue.peek}"
        if not self._values:
            return None
        return self._values[0]

    def remove(self, item: Prioritized) -> bool:
        "Implementation of L{PriorityQueue.remove}"
        try:
            self._values.remove(item)
        except ValueError:
            return False
        else:
            return True

    def __iter__(self) -> Iterator[Prioritized]:
        "Implementation of L{PriorityQueue.__iter__}"
        return iter(self._values)


_HeapIsQueue: type[PriorityQueue[int]] = Heap[int]



"""
Fritter, the Frame-Rate IndependenT TimEr tRee.

This is a library for scheduling timed events in the very near or very far
future.
"""

__version__ = "0.0.8"
"The current version of the Fritter library."



from datetime import timedelta
from json import dump

from datetype import DateTime
from fritter.drivers.datetime import DateTimeDriver, guessLocalZone
from fritter.drivers.memory import MemoryDriver
from fritter.repeat.rules.datetimes import EveryDelta

from json_identity import MyClass, registry

memoryDriver = MemoryDriver()
scheduler = registry.new(DateTimeDriver(memoryDriver))
dt = DateTime.now(guessLocalZone())
memoryDriver.advance(dt.timestamp())
myInstance = MyClass(3)
handle = scheduler.callAt(dt + timedelta(seconds=5), myInstance.later)
registry.repeatedly(
    scheduler, EveryDelta(timedelta(seconds=0.5)), myInstance.repeat
)


with open("saved-id-schedule.json", "w") as f:
    dump(registry.save(scheduler), f)



from datetime import datetime, timedelta
from typing import Callable, List, Tuple
from unittest import TestCase
from zoneinfo import ZoneInfo

from datetype import DateTime
from fritter.drivers.datetime import DateTimeDriver
from fritter.scheduler import Scheduler

from ..drivers.memory import MemoryDriver
from ..scheduler import SimpleScheduler
from ..tree import _BranchDriver, branch, timesFaster


class RecursiveTest(TestCase):
    def _oneRecursiveCall(
        self, scaleFactor: float
    ) -> List[Tuple[float, float]]:
        scheduler1: Scheduler[float, Callable[[], None]] = SimpleScheduler(
            driver := MemoryDriver()
        )
        recursive, scheduler2 = branch(scheduler1, timesFaster(scaleFactor))
        calls = []
        scheduler2.callAt(
            1.0,
            lambda: calls.append((scheduler1.now(), scheduler2.now())),
        )
        driver.advance()
        return calls

    def test_scaling(self) -> None:
        calls = self._oneRecursiveCall(1.0)
        self.assertEqual(calls, [(1.0, 1.0)])
        calls = self._oneRecursiveCall(3.0)
        self.assertEqual(calls, [(1 / 3.0, 1.0)])
        calls = self._oneRecursiveCall(1 / 3.0)
        self.assertEqual(calls, [(3.0, 1.0)])

    def test_changeScaling(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1, timesFaster(2.0))
        calls = []
        scheduler2.callAt(
            1.0,
            lambda: calls.append((scheduler1.now(), scheduler2.now())),
        )
        driver.advance(1 / 4)
        recursive.changeScale(timesFaster(4.0))
        self.assertEqual(driver.advance(), 1 / 8)
        self.assertEqual(calls, [((1 / 4) + (1 / 8), 1.0)])

    def test_datetime(self) -> None:
        scheduler1: Scheduler[DateTime[ZoneInfo], Callable[[], None]] = (
            Scheduler(DateTimeDriver(driver := MemoryDriver()))
        )
        recursive, scheduler2 = branch(scheduler1)
        TZ = ZoneInfo("Etc/UTC")
        ts = datetime(2024, 2, 9, tzinfo=TZ).timestamp()
        driver.advance(ts)
        called = False
        called2 = False

        def callme() -> None:
            nonlocal called
            called = True

        def callme2() -> None:
            nonlocal called2
            called2 = True

        scheduler2.callAt(
            DateTime.fromtimestamp(ts, TZ) + timedelta(days=1), callme
        )
        scheduler2.callAt(
            DateTime.fromtimestamp(ts, TZ) + timedelta(days=3), callme2
        )
        driver.advance(86400)
        self.assertTrue(called)
        called = False
        recursive.pause()
        driver.advance(86400)
        recursive.unpause()
        self.assertFalse(called2)
        driver.advance(86400)
        self.assertFalse(called2)
        driver.advance(86400)
        self.assertTrue(called2)
        self.assertEqual(
            scheduler1.now() - scheduler2.now(), timedelta(days=1)
        )

    def test_unscheduleNoOp(self) -> None:
        """
        Unscheduling when not scheduled is a no-op.
        """
        _BranchDriver(
            SimpleScheduler(MemoryDriver()), timesFaster(1.0), 0.0
        ).unschedule()

    def test_unpausePauseUnpause(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1, timesFaster(2))
        recursive.pause()
        self.assertEqual(scheduler2.now(), 0.0)
        driver.advance(500)
        self.assertEqual(scheduler2.now(), 0.0)
        recursive.unpause()
        self.assertEqual(scheduler2.now(), 0.0)
        driver.advance(10)
        self.assertEqual(scheduler2.now(), 20.0)
        recursive.unpause()
        self.assertEqual(scheduler2.now(), 20.0)
        driver.advance(10)
        self.assertEqual(scheduler2.now(), 40.0)

    def test_moveSooner(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1)
        calls: list[tuple[float, float]] = []
        recursive.unpause()

        recordTimestamp = timestampRecorder(calls, scheduler1, scheduler2)

        scheduler2.callAt(1.0, recordTimestamp)
        scheduler2.callAt(0.5, recordTimestamp)
        driver.advance(0.6)
        self.assertEqual(calls, [(0.6, 0.6)])

    def test_pausing(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1)
        calls = []
        scheduler2.callAt(
            1.0,
            lambda: calls.append((scheduler1.now(), scheduler2.now())),
        )
        scheduler2.callAt(
            2.0,
            lambda: calls.append((scheduler1.now(), scheduler2.now())),
        )
        self.assertEqual(calls, [])
        driver.advance(1.5)
        self.assertEqual(calls, [(1.5, 1.5)])
        calls[:] = []
        recursive.pause()
        # paused at 1.5, with 0.5 left until second call (at 2.0)
        driver.advance(2.7)
        # move to 4.2, still 0.5 left, no call yet
        self.assertEqual(calls, [])
        self.assertEqual(2.7 + 1.5, driver.now())
        self.assertEqual(1.5, scheduler2.now())
        recursive.unpause()
        driver.advance(0.5)
        self.assertEqual(2.7 + 1.5 + 0.5, driver.now())
        self.assertEqual(1.5 + 0.5, scheduler2.now())
        self.assertEqual(calls, [(2.7 + 1.5 + 0.5, 2.0)])

    def test_doubleUnpause(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        scaleFactor = 2.0
        recursive, scheduler2 = branch(scheduler1, timesFaster(scaleFactor))
        recursive.pause()
        baseTime = 1000.0
        driver.advance(baseTime)
        calls = []
        localDelta = 5.0
        scaledDelta = localDelta / scaleFactor
        scheduler2.callAt(
            scheduler2.now() + localDelta,
            lambda: calls.append((scheduler1.now(), scheduler2.now())),
        )
        recursive.unpause()
        driver.advance(1.0)
        self.assertEqual(calls, [])
        recursive.unpause()
        driver.advance(1.0)
        self.assertEqual(calls, [])
        recursive.unpause()
        driver.advance(0.5)
        self.assertEqual(calls, [(baseTime + scaledDelta, localDelta)])

    def test_idling(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1)
        calls: list[tuple[float, float]] = []
        recordTimestamp = timestampRecorder(calls, scheduler1, scheduler2)
        onlyCall = scheduler2.callAt(1.0, recordTimestamp)
        self.assertTrue(driver.isScheduled())
        onlyCall.cancel()
        self.assertFalse(driver.isScheduled())


def timestampRecorder(
    calls: list[tuple[float, float]],
    scheduler1: SimpleScheduler,
    scheduler2: SimpleScheduler,
) -> Callable[[], None]:
    def recorder() -> None:
        calls.append((scheduler1.now(), scheduler2.now()))

    return recorder



# -*- test-case-name: fritter.test.test_json -*-
"""
Serialize L{Scheduler} and the calls scheduled with it to JSON.

To use this module:

    - create a L{JSONRegistry} at module scope

    - decorate your functions and methods with L{JSONRegistry.function} and
      L{JSONRegistry.method}

    - implement L{JSONable.typeCodeForJSON}, L{JSONable.toJSON},
      L{JSONableInstance.fromJSON} as appropriate, until C{mypy} passes on your
      code

    - instantiate a L{JSONableScheduler} with an appropriate driver

    - schedule your functions and methods using it

    - save it with L{JSONRegistry.save}

    - load it later with L{JSONRegistry.load}
"""

from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import timedelta
from json import dump as save_json
from json import load as load_json
from pathlib import Path
from typing import (
    Any,
    Callable,
    Generic,
    Iterator,
    ParamSpec,
    Protocol,
    Type,
    TypeVar,
)
from zoneinfo import ZoneInfo

from datetype import DateTime, fromisoformat
from fritter.boundaries import StepsTCon

from ..boundaries import (
    Cancellable,
    RecurrenceRule,
    StepsT,
    StepsTInv,
    TimeDriver,
    RepeatingWork,
)
from ..drivers.datetime import DateTimeDriver
from ..repeat import Repeater
from ..repeat.rules.datetimes import EveryDelta
from ..scheduler import FutureCall, Scheduler

LoadContext = TypeVar("LoadContext", contravariant=True)
LoadContextCo = TypeVar("LoadContextCo", covariant=True)
"""
Each L{JSONRegistry} has its own special object passed along to
L{JSONableInstance.fromJSON}, which can be any object.  This TypeVar describes
its type.
"""
LoadContextInv = TypeVar("LoadContextInv")
"""
Like L{LoadContext} but invariant.
"""
JSONObject = dict[str, Any]
"""
A loose description of a JSON-dumpable object.
"""
# TODO: tighten JSONObject up to accurately describe valid input to
# C{json.dumps}.


class HasTypeCode(Protocol):
    """
    The L{HasTypeCode} protocol describes an object with a C{typeCodeForJSON}
    method; this is used as a bound on other types such as L{JSONableCallable},
    L{JSONableRepeatable}, and L{JSONableInstance}.
    """

    def typeCodeForJSON(self) -> str:
        """
        Type-code for the serialized JSON object to be looked up later.
        """


class JSONable(HasTypeCode, Protocol[LoadContextCo]):
    """
    Methods that allow a L{JSONRegistry} to serialize an object as JSON.
    """

    def toJSON(self, registry: JSONRegistry[LoadContextCo]) -> JSONObject:
        """
        Convert this callable to a JSON-serializable dictionary.
        """


class JSONableCallable(JSONable[LoadContextCo], Protocol):
    """
    A callable that can be serialized as JSON.
    """

    def __call__(self) -> None:
        """
        Do the work of the callable.
        """


JSONableScheduler = Scheduler[
    DateTime[ZoneInfo], JSONableCallable[LoadContextInv]
]
"""
A JSONable scheduler is a L{Scheduler} which tracks time as a L{ZoneInfo}-aware
C{datetype.DateTime} and only accepts L{JSONableCallable} objects as work to be
performed, so that they can be serialized later.

@note: The type-checker can make sure that your callalbles are json-able, but
    they can't make sure that they're registered with the correct
    L{JSONRegistry}; mixing and matching different registries will be a runtime
    error.
"""
# TODO: datetype API docs?


class JSONableRepeatable(
    JSONable[LoadContextCo], RepeatingWork[StepsTCon], Protocol
):
    """
    A callable that can be serialized as JSON, with a signature that can be
    used for repeated calls with L{JSONRegistry.repeatedly}.
    """


class MissingPersistentCall(Exception):
    """
    A serialized persistent-call handle was missing.
    """


@dataclass
class LoadProcess(Generic[LoadContextInv]):
    """
    A L{LoadProcess} collects the parameters to one top-level call to
    L{JSONRegistry.load}.
    """

    registry: JSONRegistry[LoadContextInv]
    scheduler: JSONableScheduler[LoadContextInv]
    context: LoadContextInv

    def loadFutureCall(
        self, obj: Any
    ) -> FutureCall[DateTime[ZoneInfo], JSONableCallable[LoadContextInv]]:
        callID = obj["id"]
        for call in self.scheduler._q:
            if call.id == callID:
                return call
        raise MissingPersistentCall(callID)


class JSONableInstance(JSONable[LoadContextInv], Protocol):
    """
    A class that conforms to L{JSONableInstance} can be both serialized to and
    deserialized from JSON by a L{JSONRegistry}.  L{JSONableInstance}.
    """

    @classmethod
    def typeCodeForJSON(cls) -> str:
        """
        On a L{JSONableInstance}, the C{typeCodeForJSON} method must be a
        L{classmethod}, because the class needs to be registered for
        I{de}-serialization at class-definition time, before any instances
        exist.

        @see: L{HasTypeCode}
        """

    @classmethod
    def fromJSON(
        cls: Type[JSONableSelf],
        load: LoadProcess[LoadContextInv],
        json: JSONObject,
    ) -> JSONableSelf:
        """
        Load an instance of this type from the given deserialized JSON object,
        which should be in the format returned by an instance of this class's
        L{toJSON <JSONable.toJSON>} method.
        """


JSONableSelf = TypeVar(
    "JSONableSelf",
    bound=JSONableInstance[Any],
    # See comment in JSONRegistry.method about this bound.
)
"""
TypeVar for binding C{self} on methods that want to be serialized by
L{JSONRegistry.method}.
"""

_JSONableType = TypeVar("_JSONableType", bound=HasTypeCode)
"Binding for internal generic tracking of JSONable types."

JSONRepeater = Repeater[
    DateTime[ZoneInfo],
    JSONableCallable[LoadContext],
    StepsT,
]
"A L{Repeater} that is constrained to accept only JSON-serializable types."


def _whatJSON(
    registry: JSONRegistry[LoadContextInv], what: JSONable[LoadContextInv]
) -> JSONObject:
    """
    Convert a L{JSONable} into a standard JSON-serializable object format.
    """
    return {
        "type": what.typeCodeForJSON(),
        "data": what.toJSON(registry),
    }


SomeSignature = ParamSpec("SomeSignature")
"""
A description of the parameters passed to a serializable function.
"""


@dataclass
class SerializableFunction(Generic[LoadContext, SomeSignature]):
    """
    Wrapper around a function that conforms with L{JSONable}.
    """

    original: Callable[SomeSignature, object]
    typeCode: str

    def __call__(
        self, *args: SomeSignature.args, **kwargs: SomeSignature.kwargs
    ) -> None:
        """
        Delegate invocation to the wrapped callable.
        """
        self.original(*args, **kwargs)

    def typeCodeForJSON(self) -> str:
        """
        Return the static type code supplied at construction time.
        """
        return self.typeCode

    def toJSON(self, registry: JSONRegistry[LoadContext]) -> JSONObject:
        """
        Return an empty object.
        """
        return {}


@dataclass
class JSONableBoundMethod(Generic[JSONableSelf]):
    """
    A bound method that conforms to L{JSONable} and can be serialized and
    deserialized.

    @ivar descriptor: the descriptor that this bound method is binding, created
        by L{JSONRegistry.method}.

    @ivar instance: the instance that this bound method is bound to; a
        L{JSONableInstance} of some kind.
    """

    descriptor: JSONableMethodDescriptor[JSONableSelf, Any]
    instance: JSONableSelf

    def __call__(self) -> None:
        """
        Call this bound method's function with its instance as C{self}.
        """
        self.descriptor.func(self.instance)

    def toJSON(self, registry: JSONRegistry[LoadContext]) -> JSONObject:
        """
        Convert this method's instance to a JSON-dumpable dict.
        """
        return self.instance.toJSON(registry)

    def typeCodeForJSON(self) -> str:
        """
        Combine the bound method's name with its owning class's type code.
        """
        return (
            f"{self.instance.typeCodeForJSON()}."
            f"{self.descriptor.func.__name__}"
        )


__JC: Type[JSONableCallable[object]] = JSONableBoundMethod[
    JSONableInstance[object]
]


@dataclass
class JSONableMethodDescriptor(Generic[JSONableSelf, LoadContext]):
    """
    A descriptor that can bind methods into L{JSONableBoundMethod}s.

    @see: L{JSONRegistry.method}

    @ivar registry: The L{JSONRegistry} which created this descriptor.
    @ivar func: The callable underlying method function.
    """

    registry: JSONRegistry[LoadContext]
    func: Callable[[JSONableSelf], None]

    def __set_name__(
        self, cls: Type[JSONableInstance[LoadContext]], name: str
    ) -> None:
        """
        Register the class of the decorated method when the decorator is
        registered with the class, via the protocol of L{object.__set_name__}.
        """
        self.registry._registerJSONableType(cls)

    def __get__(
        self, instance: JSONableSelf, owner: object = None
    ) -> JSONableBoundMethod[JSONableSelf]:
        """
        Bind the decorated method to an instance.
        """
        return JSONableBoundMethod(self, instance)


@dataclass
class JSONableBoundRepeatable(Generic[JSONableSelf, LoadContext, StepsTInv]):
    """
    Like L{JSONableBoundMethod}, but for repeating calls.

    @ivar descriptor: the descriptor that this bound method is binding, created
        by L{JSONRegistry.repeatMethod}.

    @ivar instance: the instance that this bound method is bound to; a
        L{JSONableInstance} of some kind.
    """

    descriptor: JSONableRepeatableDescriptor[JSONableSelf, Any, StepsTInv]
    instance: JSONableSelf

    def __call__(self, steps: StepsTInv, stopper: Cancellable) -> None:
        """
        Call this bound method's function with its instance as C{self}.
        """
        self.descriptor.func(self.instance, steps, stopper)

    def toJSON(self, registry: JSONRegistry[LoadContext]) -> JSONObject:
        """
        Convert this method's instance to a JSON-dumpable dict.
        """
        return self.instance.toJSON(registry)

    def typeCodeForJSON(self) -> str:
        """
        Combine the bound method's name with its owning class's type code.
        """
        return (
            f"{self.instance.typeCodeForJSON()}."
            f"{self.descriptor.func.__name__}"
        )


@dataclass
class JSONableRepeatableDescriptor(Generic[JSONableSelf, LoadContext, StepsT]):
    """
    A descriptor that can bind methods into L{JSONableBoundRepeatable}s.

    @see: L{JSONRegistry.repeatMethod}

    @ivar registry: The L{JSONRegistry} which created this descriptor.
    @ivar func: The callable underlying method function.
    """

    registry: JSONRegistry[LoadContext]
    func: Callable[[JSONableSelf, StepsT, Cancellable], None]

    def __set_name__(
        self, cls: Type[JSONableInstance[LoadContext]], name: str
    ) -> None:
        """
        Register the class of the decorated method when the decorator is
        registered with the class, via the protocol of L{object.__set_name__}.
        """
        self.registry._registerJSONableType(cls)

    def __get__(
        self, instance: JSONableSelf, owner: object = None
    ) -> JSONableBoundRepeatable[JSONableSelf, LoadContext, StepsT]:
        """
        Bind the decorated method to an instance.
        """
        return JSONableBoundRepeatable(self, instance)


class _JSONableMethodBinder(Protocol[LoadContextInv, StepsTInv]):
    """
    Separated description of L{JSONableMethodDescriptor.__get__} that wraps
    L{_JSONableRepeaterWrapper.repeat} to work around U{this issue
    <https://github.com/python/mypy/issues/15822>}.
    """

    def __call__(
        self,
        instance: _JSONableRepeaterWrapper[LoadContextInv, StepsTInv],
        owner: object = None,
    ) -> JSONableBoundMethod[
        _JSONableRepeaterWrapper[LoadContextInv, StepsTInv]
    ]:
        """
        Bind the method.
        """


@dataclass
class _SpecificTypeRegistration(Generic[_JSONableType]):
    """
    General container for registering a specific kind of type (instances, plain
    callables, repeated callables).

    @ivar registered: The specific type registration.
    """

    registered: dict[str, _JSONableType] = field(default_factory=dict)

    def copy(self) -> _SpecificTypeRegistration[_JSONableType]:
        """
        Copy the given L{_SpecificTypeRegistration} to derive another from it.
        """
        return _SpecificTypeRegistration(self.registered.copy())

    def add(self, item: _JSONableType) -> _JSONableType:
        """
        Add the given type to the registry, keyed by its C{typeCodeForJSON}.
        """
        self.registered[item.typeCodeForJSON()] = item
        return item

    def get(self, typeCode: str) -> _JSONableType | None:
        """
        Get a type from the registry by the result of its C{typeCodeForJSON}
        classmethod, previously registered by L{_SpecificTypeRegistration.add}.
        """
        return self.registered.get(typeCode, None)


_universal: JSONRegistry[object]
"""
Create a registry for universally-serializable types.  Right now its only
member is L{_JSONableRepeaterWrapper}, to allow for its `repeat` method to
appear as serializable to all serializable schedulers.
"""


def _copyUniversal(
    name: str,
) -> Callable[[], _SpecificTypeRegistration[_JSONableType]]:
    """
    Derive a registry type registration from the universal registry type
    registration.
    """

    def _() -> _SpecificTypeRegistration[_JSONableType]:
        result: _SpecificTypeRegistration[_JSONableType] = getattr(
            _universal, name
        )
        return result.copy()

    return _


_JSONableCallableT = TypeVar(
    "_JSONableCallableT",
    bound=JSONableCallable[Any] | JSONableRepeatable[Any, Any],
)

RRuleT = TypeVar("RRuleT", bound=RecurrenceRule[DateTime[ZoneInfo], Any])
RRuleTx = TypeVar(
    "RRuleTx",
    bound=RecurrenceRule[DateTime[ZoneInfo], Any],
    contravariant=True,
)


class RuleJSONIfier(HasTypeCode, Protocol[RRuleT]):
    """
    External serializer for a recurrence rule.
    """

    def ruleFromJSON(self, json: JSONObject) -> RRuleT:
        """
        Construct a recurrence rule from the given JSON object.
        """

    def ruleAsJSON(self, rule: RRuleT) -> JSONObject:
        """
        Serialize the recurrence rule from the given JSON object.
        """


@dataclass
class JSONRegistry(Generic[LoadContext]):
    """
    A L{JSONRegistry} maintains a set of functions and methods (and the classes
    those methods are defined on) that can scheduled against a L{Scheduler} and
    then persisted into a JSON blob to reconstitute that scheduler later.

    Use a L{JSONRegistry} by instantiatint it at module scope and then using
    its methods as decorators on functions and methods.
    """

    # TODO: implement merging multiple registries together so that objects from
    # different libraries can live in the same blob
    _functions: _SpecificTypeRegistration[JSONableCallable[LoadContext]] = (
        field(default_factory=_copyUniversal("_functions"))
    )
    _repeatable: _SpecificTypeRegistration[
        JSONableRepeatable[LoadContext, Any]
    ] = field(default_factory=_copyUniversal("_repeatable"))
    _instances: _SpecificTypeRegistration[
        Type[JSONableInstance[LoadContext]]
    ] = field(default_factory=_copyUniversal("_instances"))
    _rules: _SpecificTypeRegistration[
        RuleJSONIfier[RecurrenceRule[DateTime[ZoneInfo], Any]]
    ] = field(default_factory=_copyUniversal("_rules"))
    _ruletype2jsonifier: dict[
        type[RecurrenceRule[DateTime[ZoneInfo], object]],
        RuleJSONIfier[RecurrenceRule[DateTime[ZoneInfo], Any]],
    ] = field(default_factory=lambda: _universal._ruletype2jsonifier.copy())

    def _loadRRule(
        self, json: JSONObject
    ) -> RecurrenceRule[DateTime[ZoneInfo], Any]:
        typeCode = json["type"]
        blob = json["data"]

        if (it := self._rules.get(typeCode)) is not None:
            return it.ruleFromJSON(blob)

        raise KeyError(f"cannot interpret rule type code {repr(typeCode)}")

    def _saveRRule(
        self, rule: RecurrenceRule[DateTime[ZoneInfo], Any]
    ) -> JSONObject:
        ser = self._ruletype2jsonifier[type(rule)]
        return {
            "type": ser.typeCodeForJSON(),
            "data": ser.ruleAsJSON(rule),
        }

    def _registerRRule(
        self,
        ruleType: Type[RRuleTx],
        serializer: RuleJSONIfier[RRuleTx],
    ) -> None:
        # TODO: there's a variance problem here, I think?
        self._rules.add(serializer)  # type:ignore[arg-type]
        self._ruletype2jsonifier[ruleType] = (
            serializer  # type:ignore[assignment]
        )

    def _loadOne(
        self,
        json: JSONObject,
        which: _SpecificTypeRegistration[_JSONableCallableT],
        load: LoadProcess[LoadContext],
    ) -> _JSONableCallableT:
        """
        Convert the given JSON-dumpable dict into an object of C{_JSONableType}
        via its L{JSONableInstance.fromJSON} classmethod registered in the
        given L{_SpecificTypeRegistration}.
        """
        # inverse of _whatJSON
        typeCode = json["type"]
        blob = json["data"]

        if (it := which.get(typeCode)) is not None:
            return it

        classCode, methodName = typeCode.rsplit(".", 1)
        if (instanceType := self._instances.get(classCode)) is not None:
            # TODO: record allowable instance/method pairs so that we have some
            # confidence that the resulting type here is in fact actually
            # `JSONableType`.  probably the right way to do this is to have
            # _instances live on SpecificTypeRegistration rather than be shared
            # between both repeatable/non-repeatable types

            instance = instanceType.fromJSON(load, blob)
            result: _JSONableCallableT = getattr(instance, methodName)
            return result

        # TODO: this is for *long-term* storage of sets of scheduled events, so
        # failing the entire load like this is probably not always an
        # acceptable failure mode.  We should identify any calls that failed to
        # load, remember their uninterpreted blobs to be included again in
        # save() so we don't lose them, then communicate the failure to an
        # object passed to .load()

        raise KeyError(f"cannot interpret type code {repr(typeCode)}")

    def _repeaterToJSONable(
        self, repeater: JSONRepeater[LoadContext, StepsT]
    ) -> JSONableBoundMethod[_JSONableRepeaterWrapper[LoadContext, StepsT]]:
        """
        Convert the given L{JSONRepeater} into a method that can be serialized.
        """
        return _JSONableRepeaterWrapper(self, repeater).repeat

    def repeatedly(
        self,
        scheduler: JSONableScheduler[LoadContext],
        rule: RecurrenceRule[DateTime[ZoneInfo], StepsT],
        work: JSONableRepeatable[LoadContext, StepsT],
        reference: DateTime[ZoneInfo] | None = None,
    ) -> None:
        """
        Call C{work} repeatedly, according to C{rule}, with intervals computed
        according to the given C{reference} time (or C{scheduler}'s current
        time, if not given).

        If you want to stop the repetition, C{work} is passed a L{Cancellable}
        that you can call C{.cancel()} on to do so.

        @see: L{fritter.repeat}
        """
        Repeater.new(
            scheduler, rule, work, self._repeaterToJSONable, reference
        ).repeat()

    def function(
        self, cb: Callable[[], None]
    ) -> JSONableCallable[LoadContext]:
        """
        Mark the given 0-argument, None-returning, top-level function as
        possible to serialize within this registry.  It will be serialized by
        its fully-qualified Python name.

        Use like so::

            registry = JSONRegistry[...]()

            @registry.serializedFunction
        """
        func: JSONableCallable[LoadContext] = SerializableFunction(
            cb, cb.__qualname__
        )
        return self._functions.add(func)

    def method(
        self, method: Callable[[JSONableSelf], None]
    ) -> JSONableMethodDescriptor[JSONableSelf, LoadContext]:
        """
        Mark the given method, defined at class scope in a class complying with
        the L{JSONableInstance} protocol, as possible to serialize within this
        registry.
        """
        # I want to stipulate that the method I am taking here must have a
        # 'self' parameter whose type is *at least* as strict as
        # JSONableInstance[LoadContext] but may be stricter than that.
        # However, this would require higher-kinded typevars, in order to make
        # JSONableSelf bounded by JSONableInstance[LoadContext] rather than
        # JSONableInstance[Any]: https://github.com/python/typing/issues/548
        return JSONableMethodDescriptor(self, method)

    def repeatFunction(
        self, cb: RepeatingWork[StepsT]
    ) -> JSONableRepeatable[LoadContext, StepsT]:
        """
        Mark the given function that matches the signature of L{RepeatingWork},
        i.e. one which takes a number of steps and a L{Cancellable} to cancel
        its own repetition, as looked up via its C{__qualname__} attribute.

        @return: a function that mimics the signature of the original function,
            but also conforms to the L{JSONable} protocol.
        """
        func: JSONableRepeatable[LoadContext, StepsT] = SerializableFunction(
            cb, getattr(cb, "__qualname__")
        )
        return self._repeatable.add(func)

    def repeatMethod(
        self, repeatable: Callable[[JSONableSelf, StepsT, Cancellable], None]
    ) -> JSONableRepeatableDescriptor[JSONableSelf, LoadContext, StepsT]:
        """
        Mark the given method that matches the signature of L{RepeatingWork},
        i.e. one which takes a number of steps and a L{Cancellable} to cancel
        its own repetition, as looked up via its C{__qualname__} attribute.

        @note: The registration occurs when the class is defined, specifically
            in the C{__set_name__} hook in the returned descriptor, so if you
            are calling this in an esoteric context outside a normal class
            definition, be sure to invoke that hook as well.

        @return: a descriptor that mimics the signature of the original method.
        """
        return JSONableRepeatableDescriptor(self, repeatable)

    def _registerJSONableType(
        self, cls: Type[JSONableInstance[LoadContext]]
    ) -> None:
        """
        Mark the given class as serializable via this registry, keyed by its
        L{JSONable.typeCodeForJSON} method.  Used internally as supporting
        L{JSONRegistry.method} and L{JSONRegistry.repeatMethod}.
        """
        self._instances.add(cls)

    def load(
        self,
        runtimeDriver: TimeDriver[float],
        serializedJSON: JSONObject,
        loadContext: LoadContext,
    ) -> JSONableScheduler[LoadContext]:
        """
        Load a JSON object in the format serialized from L{JSONRegistry.save}
        and a runtime L{TimeDriver}C{[float]}, returning a
        L{JSONableScheduler}.
        """
        new: JSONableScheduler[LoadContext] = Scheduler(
            DateTimeDriver(runtimeDriver),
            counter=int(serializedJSON.get("counter", "0")),
        )
        load = LoadProcess(self, new, loadContext)
        loads = []
        for callJSON in serializedJSON["scheduledCalls"]:
            when = fromisoformat(callJSON["when"]).replace(
                tzinfo=ZoneInfo(callJSON["tz"])
            )

            # Establish the FutureCall to allow for circular reference to its
            # 'what', with a fake callable.
            fakeWhat: JSONableCallable[LoadContext]
            fakeWhat = None  # type:ignore[assignment]

            placeholder = new.callAt(when, fakeWhat)
            placeholder.id = callJSON["id"]

            loads.append((placeholder, callJSON["what"]))
            # callAt needs a way to communicate future call counters (IDs)

        for ph, wt in loads:
            ph.what = self._loadOne(wt, self._functions, load)
        return new

    def new(
        self, driver: TimeDriver[DateTime[ZoneInfo]]
    ) -> JSONableScheduler[LoadContext]:
        """
        Create a new L{JSONableScheduler} with the same type as if it had been
        loaded by this L{JSONRegistry}.
        """
        return Scheduler(driver)

    def save(self, scheduler: JSONableScheduler[LoadContext]) -> JSONObject:
        """
        Serialize the given L{JSONableScheduler} to a single L{JSONObject}
        which can be passed to L{json.dumps} or L{json.loads}.
        """
        # TODO: `self` not used yet here because we're not validating that all
        # the serializable callables here are present in this specific
        # registry, but that would be a good check to have.
        return {
            "scheduledCalls": [
                {
                    "when": item.when.replace(tzinfo=None).isoformat(),
                    "tz": item.when.tzinfo.key,
                    "what": _whatJSON(self, item.what),
                    "called": item.called,
                    "canceled": item.canceled,
                    "id": item.id,
                }
                for item in scheduler.calls()
            ],
            "counter": str(scheduler.counter),
        }

    def saveFutureCall(
        self,
        futureCall: FutureCall[
            DateTime[ZoneInfo], JSONableCallable[LoadContextInv]
        ],
    ) -> dict[str, object]:
        """
        Convert a L{FutureCall} into a JSON-serializable object.
        """
        return {"id": futureCall.id}


_universal = JSONRegistry(
    _SpecificTypeRegistration(),
    _SpecificTypeRegistration(),
    _SpecificTypeRegistration(),
    _SpecificTypeRegistration(),
    {},
)


class _EveryDeltaJSONifier:
    def typeCodeForJSON(self) -> str:
        return "fritter:fixed-delta"

    def ruleFromJSON(self, json: JSONObject) -> EveryDelta:
        return EveryDelta(timedelta(*json["delta"]))

    def ruleAsJSON(self, rule: EveryDelta) -> JSONObject:
        return {"delta": rule.delta.__reduce__()[1]}


_universal._registerRRule(EveryDelta, _EveryDeltaJSONifier())


@dataclass
class _JSONableRepeaterWrapper(Generic[LoadContext, StepsT]):
    """
    Since a L{Scheduler} can only contain C{work} of a given type, which must
    have a 0-argument, C{None}-returning signature, and L{JSONRegistry.save}
    serializes its scheduler by enumerating its future calls, this is a wrapper
    with a special method registered with I{all} L{JSONRegistry}s
    automatically, to provide a JSON-serialization format for a repeated call
    with the repeat-call signature (i.e., C{(steps, stopper)}).

    @see: L{fritter.repeat}

    @ivar jsonRegistry: The specific L{JSONRegistry} to which this
        repeater-wrapper is bound.

    @ivar repeater: The L{JSONable} repeating call that can itself be
        serialized.
    """

    jsonRegistry: JSONRegistry[LoadContext]
    repeater: JSONRepeater[LoadContext, StepsT]

    @classmethod
    def typeCodeForJSON(cls) -> str:
        """
        Return a unique type code (C{"fritter:repetition"}) for this wrapper.
        """
        return "fritter:repetition"

    @classmethod
    def fromJSON(
        cls,
        load: LoadProcess[LoadContext],
        json: JSONObject,
    ) -> _JSONableRepeaterWrapper[LoadContext, StepsT]:
        """
        Deserialize a L{_JSONableRepeaterWrapper} from a JSON-dumpable dict
        previously produced by L{_JSONableRepeaterWrapper.toJSON}.
        """
        rule: RecurrenceRule[DateTime[ZoneInfo], StepsT] = (
            load.registry._loadRRule(json["rule"])
        )
        what = json["callable"]
        one = load.registry._loadOne(what, load.registry._repeatable, load)
        ref = fromisoformat(json["ts"]).replace(tzinfo=ZoneInfo(json["tz"]))
        rep = Repeater(
            load.scheduler, rule, one, load.registry._repeaterToJSONable, ref
        )
        return cls(load.registry, rep)

    def toJSON(self, registry: JSONRegistry[object]) -> JSONObject:
        """
        Serialize this L{_JSONableRepeaterWrapper} to a JSON-dumpable dict
        suitable for deserialization with L{_JSONableRepeaterWrapper.fromJSON},
        including its time, IANA timezone identifier, rule function, and
        underlying repeating callable.
        """
        when = self.repeater.reference
        # because StepsT is parameterizable in Repeater.work, we can't make
        # Repeater.work itself be a TypeVar.
        work: JSONable[object] = self.repeater.work  # type:ignore[assignment]
        return {
            "ts": when.replace(tzinfo=None).isoformat(),
            "tz": when.tzinfo.key,
            "rule": registry._saveRRule(self.repeater.rule),
            "callable": _whatJSON(registry, work),
            # "convert": is implicitly L{registry._repeaterToJSONable}
            # "scheduler": is what's doing the serializing
        }

    @_universal.method
    def repeat(self) -> None:
        """
        This C{repeat} method is what is actually serialized, using the
        bound-method support in L{JSONRegistry}.  Internally, it is implicitly
        decorated with each registry's L{JSONRegistry.method} decorator.
        """
        self.repeater.repeat()


@contextmanager
def schedulerAtPath(
    registry: JSONRegistry[LoadContext],
    driver: TimeDriver[float],
    path: Path,
    context: LoadContext,
) -> Iterator[JSONableScheduler[LoadContext]]:
    if path.exists():
        with path.open() as rf:
            scheduler = registry.load(driver, load_json(rf), context)
    else:
        scheduler = registry.new(DateTimeDriver(driver))
    yield scheduler
    with path.open("w") as wf:
        save_json(registry.save(scheduler), wf)


__all__ = [
    "JSONableScheduler",
    "JSONObject",
    "JSONRegistry",
    "JSONableRepeatable",
    "schedulerAtPath",
]



"""
L{fritter.boundaries} describes the boundaries between different parts of the
system and its interface with your application code.  It contains L{Protocol}s,
L{TypeVar}s, and constant values, but no logic of its own.
"""

from __future__ import annotations

from typing import (
    Any,
    Callable,
    Coroutine,
    Generator,
    Iterator,
    Optional,
    Protocol,
    TypeVar,
)

import sys

if sys.version_info >= (3, 12):
    from calendar import Day
else:
    from enum import IntEnum

    class Day(IntEnum):
        MONDAY = 0
        TUESDAY = 1
        WEDNESDAY = 2
        THURSDAY = 3
        FRIDAY = 4
        SATURDAY = 5
        SUNDAY = 6


class PriorityComparable(Protocol):
    """
    Protocol describing an object that can be compared for the purposes of a
    L{PriorityQueue}.
    """

    def __lt__(self, other: Any) -> bool:
        """
        Is C{self} lower priority than C{other}?
        """

    def __le__(self, other: Any) -> bool:
        """
        Is C{self} lower I{or} the same priority as C{other}?
        """


Prioritized = TypeVar("Prioritized", bound=PriorityComparable)
"""
A TypeVar for objects that can be compared for relative priority.
"""

Yield = TypeVar("Yield", covariant=True)
"""
Yield TypeVar for generators and coroutines.
"""
Send = TypeVar("Send", contravariant=True)
"""
Send TypeVar for generators and coroutines.
"""
Return = TypeVar("Return", covariant=True)
"""
Return TypeVar for generators and coroutines.
"""


class CancellableAwaitable(Protocol[Yield, Send, Return]):
    """
    An object which can be both awaited and canceled.
    """

    def __await__(self) -> Generator[Yield, Send, Return]:
        """
        This object may be awaited.
        """

    def cancel(self) -> object:
        """
        This object may be canceled.
        """


AsyncType = TypeVar("AsyncType", bound=CancellableAwaitable[Any, Any, Any])
"TypeVar bound to a L{CancellableAwaitable}"


class PriorityQueue(Protocol[Prioritized]):
    """
    High-level specification of a priority queue.
    """

    def add(self, item: Prioritized) -> None:
        """
        Add an item to the priority queue.
        """

    def get(self) -> Optional[Prioritized]:
        """
        Consume the lowest item from the priority queue.
        """

    def peek(self) -> Optional[Prioritized]:
        """
        Examine the lowest item from the priority queue without modifying the
        queue.
        """

    def remove(self, item: Prioritized) -> bool:
        """
        Remove an item and return whether it was removed or not.
        """

    def __iter__(self) -> Iterator[Prioritized]:
        """
        Iterate all of the values in the priority queue, in an unspecified
        order.
        """


class TimeDriver(Protocol[Prioritized]):
    """
    Driver interface that allows Fritter to schedule objects onto a third party
    library.
    """

    def reschedule(
        self, newTime: Prioritized, work: Callable[[], None]
    ) -> None:
        """
        Schedule C{work} to occur at C{newTime}, removing any previous C{work}
        scheduled by prior calls to C{reschedule}.
        """

    def unschedule(self) -> None:
        """
        Remove any previously-scheduled C{work}.
        """

    def now(self) -> Prioritized:
        """
        Get the current time according to the underlying library.
        """


class Cancellable(Protocol):
    """
    An object that can be cancelled.
    """

    def cancel(self) -> object:
        """
        Cancel an operation in progress.
        """


StepsT = TypeVar("StepsT", covariant=True)
StepsTCon = TypeVar("StepsTCon", contravariant=True)
StepsTInv = TypeVar("StepsTInv")

WhenT = TypeVar("WhenT", bound=PriorityComparable)
"""
TypeVar for representing a time at which something can occur; a temporal
coordinate in a timekeeping system.
"""
WhatT = TypeVar("WhatT", bound=Callable[[], None])
"""
TypeVar for representing a unit of work that can take place within the context
of a L{Scheduler}.
"""


class RepeatingWork(Protocol[StepsTCon]):
    """
    L{RepeatingWork} is work that can be performed repeatedly in a loop.
    """

    def __call__(self, steps: StepsTCon, stopper: Cancellable) -> None:
        """
        Do the work.

        @param steps: The number of steps which have passed since the previous
            invocation.

        @param stopper: An object that you can call cancel() on to stop the
            repetition.
        """


class RecurrenceRule(Protocol[WhenT, StepsT]):
    """
    A L{RecurrenceRule} is a callable that takes a reference time and a current
    time, and computes series of steps between the current recurrence and a new
    reference time for the next call.

    Depending on the application, C{StepsT} type can either be an integer
    (i.e.: a count of the number of steps that have passed between the
    reference time and the current time) or a collection of specific previous
    step timestamps, usually a collection of C{WhenT}.
    """

    def __call__(
        self, reference: WhenT, current: WhenT
    ) -> tuple[StepsT, WhenT]:
        """
        Given a reference time and a current time, compute the steps between
        the calls and the next reference time.

        @param reference: the time at which the current invocation was
            I{scheduled} to occur; i.e. the time that the call was computed to
            have been called.

        @param current: the time at which the current invocation I{actually}
            occurred; i.e. the time that the event loop got around to actually
            calling the function.

        @note: The delta between the reference time and the current time will
            I{often} be quite small.  If a system is running actively and is
            not overloaded, then this delta will be close to zero.  However,
            there are cases (some examples: a laptop goes to sleep, then wakes
            up hours later; a program schedules a call in a database and is not
            run for several weeks) when this delta can be very large.

        @return: a 2-tuple of:

                1. I{steps}; the recurrences that were expected to have
                   occurred between C{reference} and the I{current time}.  So
                   for example, for a L{RecurrenceRule} representing a
                   once-every-5-seconds recurrence, if your reference time were
                   1.0 and your current time were 15.0, then your step count
                   should be 2, since recurrences should have occurred at 6.0
                   and 11.0.  Alternately, for a C{RecurrenceRule[float,
                   list[float]]} with the same scheduled times, C{steps} will
                   be C{[6.0, 11.0]}.

                2. I{next reference time}; time at which the next recurrence
                   I{should} occur.  In our previous example, where our
                   reference time was 1.0 and current time was 15.0, the next
                   desired time should be 16.0, since that's the next 5-second
                   recurrence after 11.0.
        """


class AsyncDriver(Protocol[AsyncType]):
    """
    An L{AsyncDriver} is an interface to a library that supports awaitables
    (such as the standard library L{asyncio}), to allow L{fritter.repeat.Async}
    to return an awaitable of a library-appropriate type.
    """

    def newWithCancel(self, cancel: Callable[[], None]) -> AsyncType:
        """
        Create a new future-ish object with the given callback to execute when
        it is cancelled.
        """

    def complete(self, asyncObj: AsyncType) -> None:
        """
        The asynchronous operation described by a previous call to this
        L{AsyncDriver}'s L{newWithCancel <AsyncDriver.newWithCancel>} method
        completed successfully; un-suspend any coroutine awaiting it.
        """

    def runAsync(self, coroutine: Coroutine[AsyncType, Any, Any]) -> AsyncType:
        """
        Run the given coroutine which awaits upon L{AsyncType}.

        @note: Whether this starts the given coroutine synchronously or waits
            until the next event-loop tick is implementation-defined.
        """



from dataclasses import dataclass
from datetime import timedelta
from zoneinfo import ZoneInfo

from datetime import tzinfo
from datetype import DateTime
from typing import TYPE_CHECKING, TypeVar

from ...boundaries import RecurrenceRule, Day

DTRule = RecurrenceRule[DateTime[ZoneInfo], int]
"""
A type alias to describe a recurrence rule function that operates on aware
datetimes.
"""

TZType = TypeVar("TZType", bound=tzinfo)


@dataclass(frozen=True)
class EveryDelta:
    """
    An L{EveryDelta} is a L{RecurrenceRule} based on a L{timedelta}, that can
    cause civil times to repeat.

    @ivar delta: the time delta between recurrences
    """

    delta: timedelta

    def __call__(
        self,
        reference: DateTime[ZoneInfo],
        current: DateTime[ZoneInfo],
    ) -> tuple[int, DateTime[ZoneInfo]]:
        """
        Compute a step count and next desired recurrence time based on a
        reference time and a current time in aware datetimes, and the timedelta
        in C{self.delta}.

        @see: L{RecurrenceRule}
        """
        count = 0
        nextDesired = reference
        while nextDesired <= current:
            count += 1
            nextDesired += self.delta
        return count, nextDesired


@dataclass
class EachWeekOn:
    """
    Repeat every week, on each weekday in the given set of C{days}, at the
    given C{hour}, C{minute}, and C{second}.
    """

    days: set[Day]
    hour: int
    minute: int
    second: int = 0

    def __call__(
        self, reference: DateTime[TZType], current: DateTime[TZType]
    ) -> tuple[list[DateTime[TZType]], DateTime[TZType]]:
        sdays = sorted([day.value for day in self.days])
        steps: list[DateTime[TZType]] = []
        refDay = reference.date().weekday()
        weekOffset = 0
        while True:
            for sday in sdays:
                daydelta = timedelta(days=(weekOffset + sday) - refDay)
                candidate = (reference + daydelta).replace(
                    hour=self.hour,
                    minute=self.minute,
                    second=self.second,
                    microsecond=0,
                )

                if candidate < reference:
                    # earlier than the reference time, let's ignore
                    continue

                if candidate <= current:
                    # after the reference time but before the current time,
                    # record as a missed step, then continue
                    steps.append(candidate)
                    continue

                # it's after the reference *and* after the current time, we're
                # done
                return steps, candidate
            weekOffset += 7


weekly: DTRule = EveryDelta(timedelta(weeks=1))
"""
Weekly datetime-based delta.
"""

daily: DTRule = EveryDelta(timedelta(days=1))
"""
Daily datetime-based rule.
"""

hourly: DTRule = EveryDelta(timedelta(hours=1))
"""
Hourly datetime-based rule.
"""

if TYPE_CHECKING:
    _isRule: RecurrenceRule[DateTime[ZoneInfo], list[DateTime[ZoneInfo]]] = (
        EachWeekOn({Day.MONDAY}, 1, 1, 1)
    )



from __future__ import annotations

from dataclasses import dataclass

from fritter.boundaries import Cancellable
from fritter.persistent.json import (
    JSONObject,
    JSONRegistry,
    LoadProcess,
)

registry = JSONRegistry[dict[str, str]]()


@dataclass
class MyClass:
    value: int

    @classmethod
    def typeCodeForJSON(cls) -> str:
        return ".".join([cls.__module__, cls.__name__])

    def toJSON(
        self, registry: JSONRegistry[dict[str, str]]
    ) -> dict[str, object]:
        return {"value": self.value}

    @classmethod
    def fromJSON(
        cls, load: LoadProcess[dict[str, str]], json: JSONObject
    ) -> MyClass:
        return cls(json["value"])

    @registry.method
    def later(self) -> None:
        print("my value is", self.value)

    @registry.repeatMethod
    def repeat(self, steps: int, stopper: Cancellable) -> None:
        print(f"performing {steps} steps at {self.value}")
        self.value += steps
        if self.value > 10:
            stopper.cancel()



from typing import Callable

from datetime import datetime, timedelta
from datetype import DateTime, aware
from zoneinfo import ZoneInfo

from fritter.boundaries import TimeDriver
from fritter.drivers.memory import MemoryDriver
from fritter.drivers.datetime import DateTimeDriver
from fritter.scheduler import Scheduler

# set up memory driver
advancer = MemoryDriver()
base: TimeDriver[float] = advancer
# set up datetime driver
TZ = ZoneInfo("US/Pacific")
dtdriver: TimeDriver[DateTime[ZoneInfo]] = DateTimeDriver(base, TZ)
# set up scheduler
scheduler = Scheduler[DateTime[ZoneInfo], Callable[[], None]](dtdriver)
# create datetime
dt = datetime(2023, 5, 5, tzinfo=TZ)
# advance to the timestamp
advancer.advance(dt.timestamp() - advancer.now())
# done advancing


# define some work
def hi() -> None:
    print("hi", scheduler.now().isoformat())


# schedule the work for 2 days in the future
scheduler.callAt(aware(dt, ZoneInfo) + timedelta(days=2), hi)
advancer.advance(2 * (60 * 60 * 24))



from __future__ import annotations

from dataclasses import dataclass

from fritter.boundaries import Cancellable
from fritter.persistent.json import JSONObject, JSONRegistry, LoadProcess

registry = JSONRegistry[dict[str, str]]()


@dataclass
class FriendReminder:
    filename: str
    current: int

    @classmethod
    def typeCodeForJSON(cls) -> str:
        return ".".join([cls.__module__, cls.__name__])

    def toJSON(self, registry: JSONRegistry[object]) -> dict[str, object]:
        return {"filename": self.filename, "current": self.current}

    @classmethod
    def fromJSON(
        cls, load: LoadProcess[dict[str, str]], json: JSONObject
    ) -> FriendReminder:
        return cls(json["filename"], json["current"])

    @registry.method
    def later(self) -> None:
        print("my value is", self.current)

    @registry.repeatMethod
    def repeat(self, steps: int, stopper: Cancellable) -> None:
        self.current += steps
        print(f"performing {steps} steps at {self.current}")
        if self.current > 10:
            stopper.cancel()



from fritter.boundaries import Cancellable
from fritter.repeat import repeatedly
from fritter.repeat.rules.seconds import EverySecond
from fritter.drivers.sleep import SleepDriver
from fritter.scheduler import Scheduler
from time import sleep

driver = SleepDriver()
start = driver.now()


def work(steps: int, stopper: Cancellable) -> None:
    elapsed = driver.now() - start
    # start slow
    if elapsed < 1.0:
        sleep(0.2)
    # end slow
    print(f"took {steps} steps at {elapsed:0.2f}")
    if elapsed >= 2.0:
        stopper.cancel()


repeatedly(Scheduler(driver), work, EverySecond(0.05))
print(f"called {driver.block()} functions, then stopped")



# -*- test-case-name: fritter.test.test_tree -*-

"""
Groups of timers that may be paused, resumed, or time-scaled together.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Generic,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
    overload,
)
from typing_extensions import Self
from .boundaries import PriorityComparable

from .scheduler import FutureCall, Scheduler

_BranchTime = TypeVar("_BranchTime", bound=PriorityComparable)
_TrunkTime = TypeVar("_TrunkTime", bound=PriorityComparable)
_TrunkDelta = TypeVar("_TrunkDelta")


class Scale(Protocol[_BranchTime, _TrunkTime, _TrunkDelta]):
    """
    A L{Scale} defines a translation between a branch (i.e., "child") time
    scale, and a trunk (i.e., "parent") time scale.
    """

    def up(self, offset: _TrunkDelta, time: _BranchTime) -> _TrunkTime:
        """
        Translate C{time} from the branch time scale into the trunk time scale.
        """

    def down(self, offset: _TrunkDelta, time: _TrunkTime) -> _BranchTime:
        """
        Translate C{time} from the trunk time scale into the branch time scale.
        """

    def shift(
        self, pauseTime: _BranchTime | None, currentTime: _TrunkTime
    ) -> _TrunkDelta:
        """
        Shift the current scale forward to incorporate
        """


DT = TypeVar("DT")


class _Deltable(PriorityComparable, Protocol[DT]):
    def __add__(self, addend: DT) -> Self: ...

    @overload
    def __sub__(self, subtrahend: DT) -> Self: ...

    @overload
    def __sub__(self, subtrahend: Self) -> DT: ...


WhenT = TypeVar("WhenT", bound=_Deltable[Any])


@dataclass
class NoScale(Generic[DT]):
    T = TypeVar("T", bound=_Deltable[DT])

    def up(self, offset: DT, time: T) -> T:
        if offset is None:
            return time
        return time + offset

    def down(self, offset: DT, time: T) -> T:
        if offset is None:
            return time
        return time - offset

    def shift(self, pauseTime: T | None, currentTime: T) -> DT:
        if pauseTime is None:
            return None  # type:ignore[return-value]
        return currentTime - pauseTime


_BranchFloat = TypeVar("_BranchFloat", bound=float)
_TrunkFloat = TypeVar("_TrunkFloat", bound=float)


@dataclass
class _FloatScale(Generic[_BranchFloat, _TrunkFloat]):
    """
    @see: L{timesFaster}
    """

    _factor: float

    """
    Amount to subtract from trunk's timestamp to get to this driver's base
    relative timestamp - in trunk's (unscaled, not branch) time-scale.  When a
    L{_BranchDriver} is created, it has a default offset of 0, which means that
    the moment '.start()' is called, that is time 0 in branch time.
    """

    _fudge: _BranchFloat = 0.0  # type:ignore[assignment]
    """
    An epsilon value computed to always be large enough that the conversion up
    to trunk and down to branch time scale will not result in time going
    backwards.
    """

    def up(self, offset: _TrunkFloat, time: _BranchFloat) -> _TrunkFloat:
        computed = (time / self._factor) + offset
        trunk: _TrunkFloat
        trunk = computed  # type:ignore[assignment]
        roundTripped: _BranchFloat = self.down(offset, trunk)
        self._fudge = _subtract(time, roundTripped)
        return trunk

    def down(
        self, offset: _TrunkFloat, trunkTime: _TrunkFloat
    ) -> _BranchFloat:
        computed = ((trunkTime - offset) * self._factor) + self._fudge
        branch: _BranchFloat = computed  # type:ignore[assignment]
        return branch

    def shift(
        self, pauseTime: _BranchFloat | None, currentTime: _TrunkFloat
    ) -> _TrunkFloat:
        delta = (pauseTime / self._factor) if pauseTime else 0.0
        trunkDelta: _TrunkFloat
        trunkDelta = delta  # type:ignore[assignment]
        return _subtract(currentTime, trunkDelta)


def timesFaster(factor: float) -> Scale[float, float, float]:
    """
    Scale a C{float} time-scale by C{factor}.  e.g., in ::

        manager, branched = branch(trunk, timesFaster(3.0))

    C{branched} will be a branch scheduler running 3 times faster than
    C{trunk}.
    """
    return _FloatScale(factor)


class BranchManager(Protocol[WhenT, _TrunkDelta]):
    """
    A L{BranchManager} controls a group of timers in a branch scheduler created
    with L{branch}; pausing the passage of time in the branch, unpausing it, or
    making its relative rate of progress faster or slower.
    """

    def changeScale(self, scale: Scale[WhenT, WhenT, _TrunkDelta]) -> None:
        """
        Change the relative scale of the time coordinate system for this branch
        and for its trunk to the new, given C{scale}.  i.e.: with a scale of
        C{timesFaster(2.0)}, that means time is running 2 times faster in this
        L{BranchManager}'s temporal coordinate system, and
        C{scheduler.callAt(3.0, X)} will run C{X} when the trunk's current
        timestamp is 1.5.
        """

    def unpause(self) -> None:
        """
        Start the branched scheduler running again.
        """

    def pause(self) -> None:
        """
        Pause the passage of time in the branched scheduler, causing its C{now}
        to stop advancing and causing any timers schedule with it via
        L{Scheduler.callAt} to stop running.
        """


@overload
def branch(
    trunk: Scheduler[WhenT, Callable[[], None]],
    scale: Scale[WhenT, WhenT, _TrunkDelta],
) -> tuple[
    BranchManager[WhenT, _TrunkDelta],
    Scheduler[WhenT, Callable[[], None]],
]: ...


@overload
def branch(trunk: Scheduler[WhenT, Callable[[], None]]) -> tuple[
    BranchManager[WhenT, WhenT],
    Scheduler[WhenT, Callable[[], None]],
]: ...


def branch(
    trunk: Scheduler[WhenT, Callable[[], None]],
    scale: Scale[WhenT, WhenT, _TrunkDelta] | None = None,
) -> tuple[
    BranchManager[WhenT, _TrunkDelta],
    Scheduler[WhenT, Callable[[], None]],
]:
    """
    Derive a branch (child) scheduler from a C{trunk} (parent) scheduler.
    """
    if scale is None:
        scale = NoScale[_TrunkDelta]()
        # scale = timesFaster(1)  # type:ignore
    assert scale is not None
    driver: _BranchDriver[WhenT, WhenT, _TrunkDelta] = _BranchDriver(
        trunk, scale, scale.shift(None, trunk.now())
    )
    driver.changeScale(scale)
    branchScheduler: Scheduler[WhenT, Callable[[], None]] = Scheduler(driver)
    driver.unpause()
    return driver, branchScheduler


_F = TypeVar("_F", bound=float)


def _subtract(someFloat: _F, other: _F) -> _F:
    return someFloat - other  # type:ignore[return-value]


@dataclass
class _BranchDriver(Generic[_TrunkTime, _BranchTime, _TrunkDelta]):
    """
    Implementation of L{TimeDriver} for L{Scheduler} that is stacked on top of
    another L{Scheduler}.
    """

    trunk: Scheduler[_TrunkTime, Callable[[], None]]
    """
    The scheduler that this driver is a branch of.
    """
    # TODO: support for a generic WhatT would be nice here, but we have our own
    # function that needs to be scheduled with the trunk scheduler, and so we
    # couldn't meaningfully integrate with a higher-level persistent scheduler.

    _scale: Scale[_BranchTime, _TrunkTime, _TrunkDelta]
    """
    How much faster the branch time coordinate system is within this scheduler?
    i.e.: with a scale factor of 2, that means time is running 2 times faster
    in this branch temporal coordinate system, and C{self.callAt(3.0, X)} will
    run C{X} when the trunk's current timestamp is 1.5.
    """
    _offset: _TrunkDelta

    _pauseTime: _BranchTime | None = None
    """
    Timestamp at which we were last paused, if we were last paused.
    """

    _scheduleWhenStarted: Optional[Tuple[_BranchTime, Callable[[], None]]] = (
        None
    )

    _call: Optional[FutureCall[_TrunkTime, Callable[[], None]]] = None
    _running: bool = False

    def reschedule(
        self, desiredTime: _BranchTime, work: Callable[[], None]
    ) -> None:
        assert (
            self._call is None or self._running
        ), f"we weren't running, call should be None not {self._call}"
        self._scheduleWhenStarted = desiredTime, work
        if not self._running:
            return

        if self._call is not None:
            self._call.cancel()

        def clearAndRun() -> None:
            self._scheduleWhenStarted = None
            self._call = None
            work()

        self._call = self.trunk.callAt(
            self._scale.up(self._offset, desiredTime), clearAndRun
        )

    def unschedule(self) -> None:
        self._scheduleWhenStarted = None
        if self._call is not None:
            self._call.cancel()
            self._call = None

    def now(self) -> _BranchTime:
        if self._running:
            return self._scale.down(self._offset, self.trunk.now())
        assert (
            self._pauseTime is not None
        ), "If a timer has been paused, _pauseTime must have been set"
        return self._pauseTime

    # implementation of 'BranchManager' interface

    def unpause(self) -> None:
        if self._running:
            return
        # shift forward the offset to skip over the time during which we were
        # paused.
        self._offset = self._scale.shift(self._pauseTime, self.trunk.now())
        self._running = True
        scheduleWhenStarted = self._scheduleWhenStarted
        if scheduleWhenStarted is not None:
            desiredTime, work = scheduleWhenStarted
            self.reschedule(desiredTime, work)

    def pause(self) -> None:
        self._pauseTime = self.now()
        self._running = False
        if self._call is not None:
            self._call.cancel()
            self._call = None

    def changeScale(
        self, newScale: Scale[_BranchTime, _TrunkTime, _TrunkDelta]
    ) -> None:
        """
        Change this recursive driver to be running at the time scale of
        C{newScale}.  i.e. C{driver.changeScale(timesFaster(3.0))} will change
        this driver's rate of time passing to be 3x faster than its trunk,
        presuming it is a float-based timer.
        """
        wasRunning = self._running
        if wasRunning:
            self.pause()
            self._scale = newScale
            self.unpause()
        else:
            self._scale = newScale


__all__ = [
    "BranchManager",
    "branch",
    "timesFaster",
    "Scale",
]



"""
Implementation of L{TimeDriver} to convert floating-point POSIX timestamps into
timezone-aware datetimes.

@note: Although at runtime this module uses L{datetime.datetime} objects, its
    type hints use the U{datetype <https://pypi.org/project/datetype/>} library
    to ensure that all values have the correct type of tzinfo.  This driver
    will I{not} work with arbitrary L{datetime.timezone} zones, I{only}
    L{zoneinfo.ZoneInfo} zones, as IANA timezone identifiers are required for
    reliable long-term serialization across DST boundaries.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable
from zoneinfo import ZoneInfo

from datetype import DateTime

from ..boundaries import TimeDriver

_PS_TZ_CMD = """\
powershell \
[\
Windows.Globalization.Calendar,\
Windows.Globalization,\
ContentType=WindowsRuntime\
]\
::New().GetTimeZone()\
"""

_guessedZone: ZoneInfo | None = None


def guessLocalZone() -> ZoneInfo:
    """
    Attempt to determine the IANA timezone identifier for the local system
    using a variety of heuristics.
    """
    global _guessedZone
    if _guessedZone is not None:
        return _guessedZone
    from os import name, readlink

    if name == "nt":
        from os import popen

        ianaID = popen(_PS_TZ_CMD).read().strip()
    else:
        path = readlink("/etc/localtime").split("/")
        ianaID = "/".join(path[path.index("zoneinfo") + 1 :])
    _guessedZone = ZoneInfo(ianaID)
    return _guessedZone


@dataclass(frozen=True)
class DateTimeDriver:
    """
    Driver based on ZoneInfo-aware datetimes.

    @ivar driver: the L{TimeDriver} that this L{DateTimeDriver} is layered on
        top of, one which represents time as a POSIX timestamp as a L{float}.

    @ivar zone: the default timezone of this L{DateTimeDriver}, the one which
        will be used for results from L{now() <DateTimeDriver.now>}.  Note that
        L{reschedule(...) <DateTimeDriver.reschedule>} will still take inputs
        in any zone; this is just the default zone for outputs.
    """

    driver: TimeDriver[float]
    zone: ZoneInfo = ZoneInfo("Etc/UTC")

    def unschedule(self) -> None:
        "Implementation of L{TimeDriver.unschedule}"
        self.driver.unschedule()

    def reschedule(
        self, newTime: DateTime[ZoneInfo], work: Callable[[], None]
    ) -> None:
        "Implementation of L{TimeDriver.reschedule}"
        self.driver.reschedule(newTime.timestamp(), work)

    def now(self) -> DateTime[ZoneInfo]:
        "Implementation of L{TimeDriver.now}"
        timestamp = self.driver.now()
        return DateTime.fromtimestamp(timestamp, self.zone)


_DriverTypeCheck: type[TimeDriver[DateTime[ZoneInfo]]] = DateTimeDriver



# -*- test-case-name: fritter.test.test_repeat -*-
"""
Schedule repeated invocations of a function, indicating how many steps have
been passed so that the repeated calls may catch up to real time to preserve
timing accuracy when timers cannot always be invoked promptly.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Coroutine,
    Generic,
    TypeVar,
)

from ..boundaries import (
    AsyncDriver,
    AsyncType,
    Cancellable,
    RepeatingWork,
    RecurrenceRule,
    StepsT,
)
from ..scheduler import Scheduler, WhatT, WhenT

StepsTInv = TypeVar("StepsTInv")


RepeatingWhatT = TypeVar("RepeatingWhatT", bound=RepeatingWork[object])
"""
A TypeVar for L{Repeater} to reference a specific type of L{RepeatingWork}.
"""


@dataclass
class Repeater(Generic[WhenT, WhatT, StepsT]):
    """
    A L{Repeater} can call a L{RepeatingWork} function repeatedly.

    @ivar scheduler: The scheduler where the work will be performed.

    @ivar rule: The L{RecurrenceRule} that defines the times at which C{work}
        will be invoked.

    @ivar work: The L{RepeatingWork} that will be performed.

    @ivar convert: Since a L{Scheduler} requires that any work provided be
        specifically of its C{WhatT} type, which may need to have some property
        associated with it beyond its function signature (for example,
        additional attributes for instrumentation or serialization), so this
        C{convert} function will convert this L{Repeater} into a 0-argument
        function that both:

            1. is of the appropriate L{WhatT} for C{scheduler}, and

            2. invokes this L{Repeater}'s C{repeat} method.

    @ivar reference: The current reference time, i.e. the time at which the
        next invocation of C{work} I{should} occur.
    """

    scheduler: Scheduler[WhenT, WhatT]
    rule: RecurrenceRule[WhenT, StepsT]
    work: RepeatingWork[StepsT]
    convert: Callable[[Repeater[WhenT, WhatT, StepsT]], WhatT]
    reference: WhenT

    @classmethod
    def new(
        cls,
        scheduler: Scheduler[WhenT, WhatT],
        rule: RecurrenceRule[WhenT, StepsT],
        work: RepeatingWork[StepsT],
        convert: Callable[[Repeater[WhenT, WhatT, StepsT]], WhatT],
        reference: WhenT | None = None,
    ) -> Repeater[WhenT, WhatT, StepsT]:
        """
        Create a L{Repeater}, filling out its reference time with the L{current
        time of the given scheduler <Scheduler.now>}, if no other time is
        provided.
        """
        if reference is None:
            reference = scheduler.now()
        return cls(scheduler, rule, work, convert, reference)

    def repeat(self) -> None:
        """
        Repeat the L{work <Repeater.work>} associated with this L{Repeater}.

        Applications should call this once, and only once, after the
        L{Repeater} is created, in order to kick off the repetition.  All
        future calls should be performed via the result of C{convert} being
        called in the scheduler, or the timing of repeated invocations will be
        incorrect.
        """
        now = self.scheduler.now()
        callIncrement, self.reference = self.rule(self.reference, now)
        callRepeat = self.convert(self)
        stopHandle = self.scheduler.callAt(self.reference, callRepeat)
        self.work(callIncrement, stopHandle)


def repeatedly(
    scheduler: Scheduler[WhenT, Callable[[], None]],
    work: RepeatingWork[StepsT],
    rule: RecurrenceRule[WhenT, StepsT],
) -> None:
    """
    Create a L{Repeater} and call its C{repeat} method.  This is a utility
    function for use when you have a simple repetition set up on a scheduler
    that accepts a baseline 0-argument callable, and does not require any
    L{conversion <Repeater.convert>}.
    """
    Repeater.new(scheduler, rule, work, lambda r: r.repeat).repeat()


@dataclass
class _AsyncStopper(Generic[AsyncType]):
    """
    An implementation of L{Cancellable} which can stop the repetition kicked
    off by L{Async.repeatedly}.
    """

    driver: AsyncDriver[AsyncType]
    result: AsyncType
    timeInProgress: Cancellable | None = None
    asyncInProgress: Cancellable | None = None
    shouldComplete: bool = True

    def cancel(self) -> None:
        if self.timeInProgress is not None:
            self.timeInProgress.cancel()
        if self.asyncInProgress is not None:
            self.asyncInProgress.cancel()
        if self.shouldComplete:
            self.driver.complete(self.result)


@dataclass
class Async(Generic[AsyncType]):
    """
    An L{Async} wraps an L{AsyncDriver} and provides an implementation of
    L{repeatedly} which:

        1. C{await}s each result from its async C{work} callable, so that no
           overlapping work will be performed if an asynchronous operation
           takes longer than the repetition interval,

        2. returns an awaitable that fires when C{stopper.cancel()} has been
           called on the C{stopper} provided to that callable, and

        3. provides a C{.cancel()} implementation on that returned awaitable
           which stops any in-progress async work and raises the appropriate
           cancellation error for your framework back to the caller.

    @ivar asyncDriver: The driver that supplies awaitables for this L{Async} to
        return.
    """

    asyncDriver: AsyncDriver[AsyncType]

    def repeatedly(
        self,
        scheduler: Scheduler[WhenT, Callable[[], None]],
        rule: RecurrenceRule[WhenT, StepsTInv],
        work: Callable[
            [StepsTInv, Cancellable],
            AsyncType | Coroutine[AsyncType, Any, Any],
        ],
    ) -> AsyncType:
        """
        Kick off a repeated call within the given scheduler, returning an
        L{AsyncType} (i.e. L{Future <asyncio.Future>}, L{Deferred
        <twisted.internet.defer.Deferred>}, or similar, as defined by
        L{Async.asyncDriver}).

        If you call C{.cancel()} on the result of this method, the repetition
        will be stopped and a cancellation error will be signaled to the
        caller.  If, instead, C{work} calls its C{.cancel()} method, then the
        result of this method will complete successfully, returning C{None} to
        a coroutine awaiting it.
        """

        cancelled = False

        def reallyCancel() -> None:
            nonlocal cancelled
            cancelled = True
            asyncStopper.shouldComplete = False
            asyncStopper.cancel()

        asyncStopper: _AsyncStopper[AsyncType] = _AsyncStopper(
            self.asyncDriver,
            self.asyncDriver.newWithCancel(reallyCancel),
        )

        def complete() -> None:
            asyncStopper.asyncInProgress = None
            if asyncStopper.timeInProgress is None and not cancelled:
                repeater.repeat()

        def kickoff(steps: StepsTInv, stopper: Cancellable) -> None:
            asyncStopper.timeInProgress = stopper
            completedSynchronously: bool = False

            async def coro() -> None:
                nonlocal completedSynchronously
                try:
                    await work(steps, asyncStopper)
                finally:
                    if asyncStopper.asyncInProgress is None:
                        completedSynchronously = True
                    else:
                        complete()

            asyncStopper.asyncInProgress = self.asyncDriver.runAsync(coro())
            if completedSynchronously:
                complete()

        def whenReady() -> None:
            asyncStopper.timeInProgress = None
            if asyncStopper.asyncInProgress is None:
                repeater.repeat()

        repeater = Repeater.new(scheduler, rule, kickoff, lambda r: whenReady)
        repeater.repeat()

        return asyncStopper.result



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from zoneinfo import ZoneInfo

from datetype import aware
from fritter.drivers.datetime import DateTimeDriver, guessLocalZone
from fritter.drivers.memory import MemoryDriver
from fritter.persistent.json import (
    JSONObject,
    JSONRegistry,
    LoadProcess,
)

registry = JSONRegistry[dict[str, str]]()


@dataclass
class MyClass:
    value: int

    @classmethod
    def typeCodeForJSON(cls) -> str:
        return ".".join([cls.__module__, cls.__name__])

    def toJSON(self, registry: JSONRegistry[object]) -> dict[str, object]:
        return {"value": self.value}

    @classmethod
    def fromJSON(
        cls, load: LoadProcess[dict[str, str]], json: JSONObject
    ) -> MyClass:
        return cls(json["value"])

    @registry.method
    def later(self) -> None:
        print("my value is", self.value)


memoryDriver = MemoryDriver()
scheduler = registry.new(DateTimeDriver(memoryDriver))
dt = aware(
    datetime(
        2023,
        7,
        21,
        1,
        1,
        1,
        tzinfo=guessLocalZone(),
    ),
    ZoneInfo,
)

handle = scheduler.callAt(dt, MyClass(3).later)
myInstance = MyClass(3)
from json import dumps, loads

dump = dumps(registry.save(scheduler))
print(dump)
mem2 = MemoryDriver()
loaded = registry.load(mem2, loads(dump), {})
mem2.advance(dt.timestamp())



from datetime import timedelta
from json import dump

from datetype import DateTime
from fritter.drivers.datetime import DateTimeDriver, guessLocalZone
from fritter.drivers.memory import MemoryDriver

from json_instance import MyClass, registry

memoryDriver = MemoryDriver()
scheduler = registry.new(DateTimeDriver(memoryDriver))
dt = DateTime.now(guessLocalZone())
handle = scheduler.callAt(dt + timedelta(seconds=5), MyClass(3).later)
myInstance = MyClass(3)


with open("saved-schedule.json", "w") as f:
    dump(registry.save(scheduler), f)



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from json import dump, load
from pathlib import Path
from typing import Any, Iterator
from zoneinfo import ZoneInfo

from datetype import DateTime, aware, fromisoformat

from fritter.boundaries import Cancellable, TimeDriver
from fritter.drivers.datetime import DateTimeDriver, guessLocalZone
from fritter.drivers.memory import MemoryDriver
from fritter.drivers.sleep import SleepDriver
from fritter.persistent.json import (
    JSONObject,
    JSONRegistry,
    JSONableScheduler,
    LoadProcess,
)
from fritter.repeat.rules.datetimes import weekly


registry: JSONRegistry[FriendList] = JSONRegistry()


@dataclass
class FriendList:
    friendsByName: dict[str, Friend] = field(default_factory=dict)

    def save(self, scheduler: JSONableScheduler[FriendList]) -> dict[str, Any]:
        return {
            "friends": [
                friend.asFriendListJSON()
                for friend in self.friendsByName.values()
            ],
            "scheduler": registry.save(scheduler),
        }

    @classmethod
    def load(
        cls, driver: TimeDriver[float], json: dict[str, Any]
    ) -> tuple[FriendList, JSONableScheduler[FriendList]]:
        friendsByName = {}
        for friendJSON in json["friends"]:
            friend = Friend.fromFriendListJSON(friendJSON)
            friendsByName[friend.name] = friend
        self = cls(friendsByName)
        return self, registry.load(driver, json["scheduler"], self)

    @classmethod
    def typeCodeForJSON(cls) -> str:
        return "friend-list"

    def toJSON(self, registry: JSONRegistry[object]) -> dict[str, object]:
        return {}

    @classmethod
    def fromJSON(
        cls, load: LoadProcess[FriendList], json: JSONObject
    ) -> FriendList:
        return load.context

    @registry.repeatMethod
    def getInTouch(self, steps: int, stopper: Cancellable) -> None:
        byContact = sorted(
            self.friendsByName.values(),
            key=lambda f: f.lastContact,
        )
        if (friend := next(iter(byContact), None)) is None:
            print("nobody to get in touch with right now")
            return
        print("maybe you should get in touch with", friend.name)
        friend.lastContact = now()

    def add(
        self,
        name: str,
        birthdayDay: int,
        birthdayMonth: int,
        lastContact: DateTime[ZoneInfo],
    ) -> Friend:
        f = self.friendsByName[name] = Friend(
            name, birthdayDay, birthdayMonth, lastContact
        )
        return f

    @classmethod
    def new(
        cls, driver: TimeDriver[float]
    ) -> tuple[FriendList, JSONableScheduler[FriendList]]:
        self = cls()
        scheduler = registry.new(DateTimeDriver(driver))
        registry.repeatedly(scheduler, weekly, self.getInTouch)
        return self, scheduler


TZ = guessLocalZone()


def now() -> DateTime[ZoneInfo]:
    return DateTime.now(TZ)


# friend-class
@dataclass
class Friend:
    name: str
    birthdayDay: int
    birthdayMonth: int
    lastContact: DateTime[ZoneInfo]
    # end-friend-fields

    def asFriendListJSON(self) -> dict[str, object]:
        return {
            "name": self.name,
            "birthdayDay": self.birthdayDay,
            "birthdayMonth": self.birthdayMonth,
            "lastContact": {
                "ts": self.lastContact.replace(tzinfo=None).isoformat(),
                "tz": self.lastContact.tzinfo.key,
            },
        }

    @classmethod
    def fromFriendListJSON(cls, json: dict[str, Any]) -> Friend:
        return cls(
            json["name"],
            json["birthdayDay"],
            json["birthdayMonth"],
            fromisoformat(json["lastContact"]["ts"]).replace(
                tzinfo=ZoneInfo(json["lastContact"]["tz"])
            ),
        )

    @classmethod
    def typeCodeForJSON(cls) -> str:
        return "friend"

    def toJSON(self, registry: JSONRegistry[object]) -> dict[str, object]:
        return {"name": self.name}

    @classmethod
    def fromJSON(
        cls, load: LoadProcess[FriendList], json: JSONObject
    ) -> Friend:
        return load.context.friendsByName[json["name"]]


saved = Path("friend-schedule.json")


@contextmanager
def listLoaded(driver: TimeDriver[float]) -> Iterator[FriendList]:
    if saved.exists():
        with saved.open() as f:
            friendList, scheduler = FriendList.load(driver, load(f))
    else:
        friendList, scheduler = FriendList.new(driver)
    yield friendList
    blob = friendList.save(scheduler)
    with saved.open("w") as f:
        dump(blob, f)


@contextmanager
def realTimeList() -> Iterator[FriendList]:
    driver = SleepDriver()
    with listLoaded(driver) as fl:
        driver.block(1.0)
        yield fl


@dataclass
class Storyteller:
    currentTime: DateTime[ZoneInfo]

    @contextmanager
    def fakeTimeList(self, toAdvance: float) -> Iterator[FriendList]:
        driver = MemoryDriver(self.currentTime.timestamp())
        with listLoaded(driver) as fl:
            yield fl
            driver.advance(toAdvance)
        self.currentTime = DateTime.fromtimestamp(driver.now(), tz=TZ)


TZ = guessLocalZone()


def story() -> None:
    start = aware(datetime(2023, 11, 1, 9, 0, 0, tzinfo=TZ), ZoneInfo)

    print("day 1")
    st = Storyteller(start)
    with st.fakeTimeList(0) as fl1:
        fl1.add("alice", 12, 1, start)
        fl1.add("bob", 12, 15, start)

    print("day 10")
    with st.fakeTimeList(timedelta(days=10).total_seconds()):
        pass

    print("day 30")
    with st.fakeTimeList(timedelta(days=20).total_seconds()):
        pass

    print("day 40")
    with st.fakeTimeList(timedelta(days=10).total_seconds()):
        pass

    print("day 90")
    with st.fakeTimeList(timedelta(days=50).total_seconds()):
        pass
    print("done")
if __name__ == "__main__":
    story()



from unittest import TestCase

from ..drivers.sleep import SleepDriver
from ..scheduler import SimpleScheduler


class TestSleeping(TestCase):
    def test_sleep(self) -> None:
        sleeps = []
        current = 0.0

        def sleep(duration: float) -> None:
            nonlocal current
            sleeps.append(duration)
            current += duration

        def time() -> float:
            return current

        driver = SleepDriver(sleep=sleep, time=time)
        scheduler = SimpleScheduler(driver)

        threeCalledAt = None
        sevenCalledAt = None

        def three() -> None:
            nonlocal threeCalledAt
            threeCalledAt = driver.now()

        def seven() -> None:
            nonlocal sevenCalledAt
            sevenCalledAt = driver.now()

        scheduler.callAt(3.0, three)
        scheduler.callAt(7.0, seven)
        driver.block()

        self.assertEqual(threeCalledAt, 3.0)
        self.assertEqual(sevenCalledAt, 7.0)
        self.assertEqual(sleeps, [3.0, 4.0])

    def test_unschedule(self) -> None:
        sleeps = []
        current = 0.0

        def sleep(duration: float) -> None:
            nonlocal current
            sleeps.append(duration)
            current += duration

        def time() -> float:
            return current

        driver = SleepDriver(sleep=sleep, time=time)
        scheduler = SimpleScheduler(driver)

        times = 0

        def once() -> None:
            nonlocal times
            times += 1

        scheduler.callAt(1, once)
        two = scheduler.callAt(2, once)
        driver.block(1.5)
        self.assertEqual(times, 1)
        two.cancel()
        driver.block(1.5)
        self.assertEqual(times, 1)

    def test_timeout(self) -> None:
        current = 0.0
        sleeps = []
        calls = []

        def sleep(duration: float) -> None:
            nonlocal current
            sleeps.append(duration)
            current += duration

        def time() -> float:
            return current

        driver = SleepDriver(sleep, time)

        def hello() -> None:
            calls.append(1)

        driver.reschedule(1.0, hello)
        result = driver.block(0.4)

        self.assertEqual(sleeps, [0.4])
        self.assertEqual(result, 0)
        self.assertEqual(current, 0.4)
        self.assertEqual(calls, [])

        result = driver.block(0.6)
        self.assertEqual(sleeps, [0.4, 0.6])
        self.assertEqual(result, 1)
        self.assertEqual(current, 1.0)
        self.assertEqual(calls, [1])

        result = driver.block(0.7)
        self.assertEqual(result, 0)
        self.assertEqual(current, 1.0)
        self.assertEqual(calls, [1])
        self.assertEqual(sleeps, [0.4, 0.6])



from unittest import TestCase

from ..drivers.memory import MemoryDriver


class MemoryDriverTests(TestCase):
    def test_advance(self) -> None:
        driver = MemoryDriver()
        work = []
        self.assertEqual(driver.isScheduled(), False)
        driver.reschedule(3.5, lambda: work.append(driver.now()))
        self.assertEqual(driver.isScheduled(), True)
        driver.advance(1)
        self.assertEqual(driver.now(), 1.0)
        self.assertEqual(work, [])
        self.assertEqual(driver.advance(), 2.5)
        self.assertEqual(driver.now(), 3.5)
        self.assertEqual(driver.isScheduled(), False)
        self.assertEqual(driver.advance(), None)

    def test_noBackwards(self) -> None:
        driver = MemoryDriver()
        count = 0

        def work() -> None:
            nonlocal count
            count += 1
            driver.reschedule(0, work)

        driver.reschedule(0, work)
        driver.advance()
        self.assertEqual(count, 1)
        self.assertGreater(driver.now(), 0.0)
        self.assertLess(driver.now(), 1e-20)



# setup
from fritter.drivers.memory import MemoryDriver
from fritter.scheduler import SimpleScheduler
from fritter.tree import branch, timesFaster

driver = MemoryDriver()
trunk = SimpleScheduler(driver)
rate = 3.0
manager, branched = branch(trunk, timesFaster(rate))
# end setup


# showfunc
def loop(scheduler: SimpleScheduler, name: str, interval: float = 1.0) -> None:
    def _() -> None:
        print(name)
        scheduler.callAt(scheduler.now() + interval, _)

    _()


# end showfunc

# loops
loop(trunk, "trunk", 1.0)
loop(branched, "branch", 1.0)
# work
for again in range(10):
    driver.advance()
    rate += 1
    manager.changeScale(timesFaster(rate))
    print(f"time: trunk={trunk.now()} branch={branched.now()}")



from dataclasses import dataclass

from typing import TYPE_CHECKING


@dataclass
class EverySecond:
    """
    An L{EverySecond} is a L{RecurrenceRule} based on a L{float} timestamp,
    that can repeat work on a physical-time interval.

    @ivar seconds: the number of seconds between recurrences.
    """

    seconds: float

    def __call__(self, reference: float, current: float) -> tuple[int, float]:
        """
        Compute a step count and next desired recurrence time based on a
        reference time and a current time in floating-point POSIX timestamps
        like those returned by L{time.time} and L{datetime.datetime.timestamp},
        and number of seconds in L{self.seconds <EverySecond.seconds>}.

        @see: L{RecurrenceRule}
        """
        elapsed = current - reference
        count, remainder = divmod(elapsed, self.seconds)
        return int(count) + 1, current + (self.seconds - remainder)


if TYPE_CHECKING:
    from ...boundaries import RecurrenceRule

    _isRule: RecurrenceRule[float, int] = EverySecond(1.0)



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timedelta
from json import dumps, loads
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, Type
from unittest import TestCase
from zoneinfo import ZoneInfo

from datetype import DateTime, aware
from fritter.persistent.json import schedulerAtPath

from ..boundaries import Cancellable, TimeDriver
from ..drivers.datetime import DateTimeDriver
from ..drivers.memory import MemoryDriver
from ..persistent.json import (
    JSONableCallable,
    JSONableInstance,
    JSONableScheduler,
    JSONObject,
    JSONRegistry,
    LoadProcess,
    MissingPersistentCall,
)
from ..repeat.rules.datetimes import daily
from ..scheduler import FutureCall


@dataclass
class RegInfo:
    calls: list[str]
    identityMap: dict[str, Any] = field(default_factory=dict)


registry = JSONRegistry[RegInfo]()
emptyRegistry = JSONRegistry[RegInfo]()
PT = ZoneInfo(key="America/Los_Angeles")

calls = []


@registry.function
def call1() -> None:
    calls.append("hello")


@registry.function
def call2() -> None:
    calls.append("goodbye")


@registry.repeatFunction
def repeatable(steps: int, stopper: Cancellable) -> None:
    calls.append(f"repeatable {steps}")


@dataclass
class InstanceWithMethods:
    value: str
    info: RegInfo
    calls: int = 0
    stoppers: list[Cancellable] = field(default_factory=list)

    @classmethod
    def typeCodeForJSON(self) -> str:
        return "instanceWithMethods"

    @classmethod
    def fromJSON(
        cls, load: LoadProcess[RegInfo], json: JSONObject
    ) -> InstanceWithMethods:
        key = json["identity"]
        if key in load.context.identityMap:
            load.context.calls.append("InstanceWithMethods.fromJSON (cached)")
            self: InstanceWithMethods = load.context.identityMap[key]
            return self
        load.context.calls.append("InstanceWithMethods.fromJSON")
        new = cls(json["value"], load.context)
        load.context.identityMap[key] = new
        return new

    def toJSON(self, registry: JSONRegistry[RegInfo]) -> dict[str, object]:
        return {
            "value": self.value,
            "identity": id(self),
        }

    @registry.method
    def method1(self) -> None:
        self.info.calls.append(f"{self.value}/method1")

    @registry.method
    def method2(self) -> None:
        self.info.calls.append(f"{self.value}/method2")

    @registry.repeatMethod
    def repeatMethod(self, steps: int, stopper: Cancellable) -> None:
        self.calls += 1
        self.stoppers.append(stopper)
        self.info.calls.append(
            f"repeatMethod {steps} {self.value=} {self.calls=}"
        )


Handle = FutureCall[DateTime[ZoneInfo], JSONableCallable[RegInfo]]


@dataclass
class Stoppable:
    runcall: Handle | None = None
    stopcall: Handle | None = None
    ran: bool = False

    def scheduleme(self, scheduler: JSONableScheduler[RegInfo]) -> None:
        """
        Schedule 'runme' to run 2 seconds in the future, but 'stopme' to run 1
        second in the future.
        """
        now = scheduler.now()
        self.runcall = scheduler.callAt(now + timedelta(seconds=2), self.runme)
        self.stopcall = scheduler.callAt(
            now + timedelta(seconds=1), self.stopme
        )

    @classmethod
    def typeCodeForJSON(self) -> str:
        return "stoppable"

    def toJSON(self, registry: JSONRegistry[RegInfo]) -> dict[str, object]:
        def save(it: Handle | None) -> object:
            return registry.saveFutureCall(it) if it is not None else it

        return {
            "runcall": save(self.runcall),
            "stopcall": save(self.stopcall),
            "ran": self.ran,
            "id": id(self),
        }

    @classmethod
    def fromJSON(
        cls, load: LoadProcess[RegInfo], json: JSONObject
    ) -> Stoppable:
        if json["id"] in load.context.identityMap:
            result: Stoppable = load.context.identityMap[json["id"]]
            return result

        def get(
            name: str,
        ) -> Handle | None:
            it = json[name]
            return it if it is None else load.loadFutureCall(it)

        self = cls(
            runcall=get("runcall"), stopcall=get("stopcall"), ran=json["ran"]
        )
        # leave it there for the test to pick up
        load.context.identityMap[json["id"]] = self
        return self

    @registry.method
    def stopme(self) -> None:
        assert self.runcall is not None
        self.stopcall = None
        self.runcall.cancel()
        self.runcall = None

    @registry.method
    def runme(self) -> None:
        self.ran = True
        self.runcall = None


stp: Type[JSONableInstance[RegInfo]] = Stoppable


def jsonScheduler(driver: TimeDriver[float]) -> JSONableScheduler[RegInfo]:
    return registry.new(DateTimeDriver(driver))


class PersistentSchedulerTests(TestCase):
    def tearDown(self) -> None:
        del calls[:]

    def test_scheduleRunSaveRun(self) -> None:
        """
        Test scheduling module-level functions and instance methods.
        """
        memoryDriver = MemoryDriver()
        scheduler = jsonScheduler(memoryDriver)
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        dt2 = aware(
            datetime(2023, 7, 22, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        ri0 = RegInfo([])
        iwm = InstanceWithMethods("test_scheduleRunSaveRun value", ri0)
        scheduler.callAt(dt, call1)
        scheduler.callAt(dt, iwm.method1)
        scheduler.callAt(dt2, call2)
        scheduler.callAt(dt2, iwm.method2)
        memoryDriver.advance(dt.timestamp() + 1)
        self.assertEqual(calls, ["hello"])
        del calls[:]
        saved = registry.save(scheduler)
        memory2 = MemoryDriver()
        ri = RegInfo([])
        registry.load(memory2, saved, ri)
        memory2.advance(dt2.timestamp() + 1)
        self.assertEqual(calls, ["goodbye"])
        self.assertEqual(ri0.calls, ["test_scheduleRunSaveRun value/method1"])
        self.assertEqual(
            ri.calls,
            [
                "InstanceWithMethods.fromJSON",
                "test_scheduleRunSaveRun value/method2",
            ],
        )

    def test_persistCancellers(self) -> None:
        """
        scheduled instance methods ought to be able to save handles to other
        instances and stuff
        """
        memoryDriver = MemoryDriver()
        scheduler = jsonScheduler(memoryDriver)
        dt = aware(datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT), ZoneInfo)
        memoryDriver.advance(dt.timestamp() + 1)
        s = Stoppable()
        self.assertEqual(s.ran, False)
        s.runme()
        self.assertEqual(s.ran, True)

        s = Stoppable()
        s.scheduleme(scheduler)
        jsonobj = dumps(registry.save(scheduler))
        saved = loads(jsonobj)
        memory2 = MemoryDriver()
        ri = RegInfo([])
        registry.load(memory2, saved, ri)
        [(name, loadedStoppable)] = ri.identityMap.items()
        assert isinstance(loadedStoppable, Stoppable)
        self.assertEqual(loadedStoppable.ran, False)
        self.assertIsNot(loadedStoppable.runcall, None)
        memory2.advance(dt.timestamp() + 3.0)
        self.assertEqual(loadedStoppable.ran, False)
        self.assertIs(loadedStoppable.runcall, None)

    def test_schedulerAtPath(self) -> None:
        ri0 = RegInfo([])
        iwm = InstanceWithMethods("A", ri0)
        mem = MemoryDriver()
        p = Path(mkdtemp()) / "scheduler.json"
        ts = datetime(2024, 2, 1, tzinfo=ZoneInfo("Etc/UTC")).timestamp()
        mem.advance(ts)
        aw = aware(datetime(2024, 2, 2, tzinfo=ZoneInfo("Etc/UTC")), ZoneInfo)
        with schedulerAtPath(registry, mem, p, ri0) as sched1:
            sched1.callAt(aw, iwm.method1)
        ri1 = RegInfo([])
        mem2 = MemoryDriver()
        with schedulerAtPath(registry, mem2, p, ri1):
            mem2.advance()
        self.assertEqual(
            ri1.calls, ["InstanceWithMethods.fromJSON", "A/method1"]
        )

    def test_noSuchCallID(self) -> None:
        mem = MemoryDriver()
        with self.assertRaises(MissingPersistentCall) as raised:
            registry.load(
                mem,
                {
                    "scheduledCalls": [
                        {
                            "when": "2023-07-21T08:01:03",
                            "tz": "Etc/UTC",
                            "what": {
                                "type": "stoppable.stopme",
                                "data": {
                                    "runcall": {"id": 7},
                                    "stopcall": {"id": 2},
                                    "ran": False,
                                    "id": 4411099664,
                                },
                            },
                            "called": False,
                            "canceled": False,
                            "id": 2,
                        }
                    ],
                    "counter": "2",
                },
                RegInfo([]),
            )
        self.assertEqual(raised.exception.args[0], 7)

    def test_idling(self) -> None:
        memoryDriver = MemoryDriver()
        scheduler = jsonScheduler(memoryDriver)
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        handle = scheduler.callAt(dt, call1)
        self.assertEqual(memoryDriver.isScheduled(), True)
        handle.cancel()
        self.assertEqual(memoryDriver.isScheduled(), False)
        memoryDriver.advance(dt.timestamp() + 1)
        self.assertEqual(calls, [])

    def test_emptyScheduler(self) -> None:
        memory = MemoryDriver()
        registry.load(memory, {"scheduledCalls": []}, RegInfo([]))
        self.assertEqual(memory.isScheduled(), False)

    def test_repeatable(self) -> None:
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        memoryDriver = MemoryDriver()
        memoryDriver.advance(dt.timestamp())
        scheduler = jsonScheduler(memoryDriver)
        registry.repeatedly(scheduler, daily, repeatable, dt)
        self.assertEqual(calls, ["repeatable 1"])
        del calls[:]

        def days(n: int) -> float:
            return 60 * 60 * 24 * n

        memoryDriver.advance(days(3))
        self.assertEqual(calls, ["repeatable 3"])
        del calls[:]

        newInfo = RegInfo([])
        mem2 = MemoryDriver()
        mem2.advance(dt.timestamp())
        mem2.advance(days(7))
        self.assertEqual(mem2.isScheduled(), False)
        registry.load(mem2, loads(dumps(registry.save(scheduler))), newInfo)
        self.assertEqual(mem2.isScheduled(), True)
        amount = mem2.advance()
        assert amount is not None
        self.assertLess(amount, 0.0001)
        self.assertEqual(calls, ["repeatable 4"])

    def test_repeatLoadError(self) -> None:
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        memoryDriver = MemoryDriver()
        memoryDriver.advance(dt.timestamp())
        oneCall = {
            "when": "2023-07-22T08:01:01",
            "tz": "Etc/UTC",
            "what": {
                "type": "fritter:repetition.repeat",
                "data": {
                    "ts": "2023-07-22T08:01:01",
                    "tz": "Etc/UTC",
                    "rule": {
                        "type": "incorrect rule type",
                        "data": {"delta": (1, 0, 0)},
                    },
                    "callable": {
                        "type": "instanceWithMethods.repeatMethod",
                        "data": {
                            "value": "sample",
                            "identity": 4335201296,
                        },
                    },
                },
            },
            "called": False,
            "canceled": False,
            "id": 1,
        }
        with self.assertRaises(KeyError) as ke:
            registry.load(
                memoryDriver,
                {
                    "scheduledCalls": [oneCall],
                    "counter": "1",
                },
                RegInfo([]),
            )
        self.assertEqual(
            str(ke.exception),
            repr("cannot interpret rule type code 'incorrect rule type'"),
        )

    def test_repeatableMethod(self) -> None:
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        memoryDriver = MemoryDriver()
        memoryDriver.advance(dt.timestamp())
        scheduler = jsonScheduler(memoryDriver)
        info = RegInfo([])
        inst = InstanceWithMethods("sample", info)
        method = inst.repeatMethod
        shared = InstanceWithMethods("shared", info)
        registry.repeatedly(scheduler, daily, method)
        registry.repeatedly(scheduler, daily, shared.repeatMethod)
        registry.repeatedly(scheduler, daily, shared.repeatMethod)
        self.assertEqual(
            info.calls,
            [
                "repeatMethod 1 self.value='sample' self.calls=1",
                "repeatMethod 1 self.value='shared' self.calls=1",
                "repeatMethod 1 self.value='shared' self.calls=2",
            ],
        )
        del info.calls[:]

        def days(n: int) -> float:
            return 60 * 60 * 24 * n

        memoryDriver.advance(days(3))
        self.assertEqual(
            info.calls,
            [
                "repeatMethod 3 self.value='sample' self.calls=2",
                "repeatMethod 3 self.value='shared' self.calls=3",
                "repeatMethod 3 self.value='shared' self.calls=4",
            ],
        )

        newInfo = RegInfo([])
        newNewInfo = RegInfo([])

        def atTimeDriver() -> MemoryDriver:
            x = MemoryDriver()
            x.advance(dt.timestamp())
            x.advance(days(7))
            return x

        mem2 = atTimeDriver()
        mem3 = atTimeDriver()

        self.assertEqual(mem2.isScheduled(), False)
        persistent = dumps(registry.save(scheduler))
        loadedScheduler = registry.load(mem2, loads(persistent), newInfo)
        repersistent = dumps(registry.save(loadedScheduler))
        registry.load(mem3, loads(repersistent), newNewInfo)
        loaded = loads(persistent)
        with self.assertRaises(KeyError):
            # TODO: allow for better error handling that doesn't just blow up
            # on the type code lookup failure
            emptyRegistry.load(MemoryDriver(), loaded, newInfo)
        self.assertEqual(mem2.isScheduled(), True)
        mem2.advance()
        expectedCalls = [
            "InstanceWithMethods.fromJSON",
            "InstanceWithMethods.fromJSON",
            "InstanceWithMethods.fromJSON (cached)",
            "repeatMethod 4 self.value='sample' self.calls=1",
            "repeatMethod 4 self.value='shared' self.calls=1",
            "repeatMethod 4 self.value='shared' self.calls=2",
        ]
        self.assertEqual(newInfo.calls, expectedCalls)
        self.assertEqual(mem3.isScheduled(), True)
        mem3.advance()
        self.assertEqual(newNewInfo.calls, expectedCalls)

        # round trip:



from unittest import TestCase
from zoneinfo import ZoneInfo

from ..drivers.datetime import guessLocalZone


class ZoneSmokeTest(TestCase):
    def test_guessLocalZone(self) -> None:
        first = guessLocalZone()
        self.assertIsInstance(first, ZoneInfo)
        second = guessLocalZone()
        self.assertIs(first, second)
        print("Guessed Zone:", first.key)



from datetime import datetime
from itertools import chain
from typing import Any, Callable
from unittest import TestCase
from zoneinfo import ZoneInfo

from datetype import DateTime
from twisted.internet.defer import CancelledError, Deferred, succeed

from ..boundaries import Cancellable, Day, RecurrenceRule
from ..drivers.datetime import DateTimeDriver
from ..drivers.memory import MemoryDriver
from ..drivers.twisted import TwistedAsyncDriver
from ..repeat import Async, repeatedly
from ..repeat.rules.datetimes import EachWeekOn
from ..repeat.rules.seconds import EverySecond
from ..scheduler import Scheduler


class RepeatTestCase(TestCase):
    def test_synchronous(self) -> None:
        mem = MemoryDriver()
        calls = []

        def work(steps: int, stopper: Cancellable) -> None:
            now = mem.now()
            if mem.now() >= 10.0:
                stopper.cancel()
            calls.append((steps, now))

        repeatedly(Scheduler(mem), work, EverySecond(5))

        self.assertTrue(mem.isScheduled())
        self.assertEqual(calls, [(1, 0.0)])
        calls = []
        mem.advance()
        self.assertTrue(mem.isScheduled())
        self.assertEqual(calls, [(1, 5.0)])
        calls = []
        mem.advance()
        self.assertFalse(mem.isScheduled())
        self.assertEqual(calls, [(1, 10.0)])

    def test_repeatEveryIntervalInSeconds(self) -> None:
        tad = TwistedAsyncDriver()
        mem = MemoryDriver()
        count = 0
        calls = []
        event: Deferred[None] = succeed(None)

        async def tick(times: int) -> None:
            nonlocal count, event
            calls.append(f"before {count} ({times})")
            event = Deferred()
            await event
            calls.append(f"after {count}")
            count += 1

        Async(tad).repeatedly(
            Scheduler(mem),
            EverySecond(15),
            lambda times, stopper: tick(times),
        )

        self.assertEqual(calls, ["before 0 (1)"])
        event.callback(None)
        self.assertEqual(calls, ["before 0 (1)", "after 0"])
        del calls[:]
        mem.advance(15.3)
        self.assertEqual(calls, ["before 1 (1)"])
        del calls[:]
        # async operation takes 45 seconds. it's now 60.3.
        mem.advance(45.0)
        self.assertEqual(calls, [])
        event.callback(None)
        # catch-up call is immediately scheduled
        self.assertEqual(calls, ["after 1", "before 2 (3)"])
        del calls[:]
        event.callback(None)
        self.assertEqual(calls, ["after 2"])

    def test_complete(self) -> None:
        count = 0
        threshold = 3

        def step(steps: int, stopper: Cancellable) -> Deferred[None]:
            nonlocal count
            count += steps
            if count >= threshold:
                stopper.cancel()
            return succeed(None)

        tad = TwistedAsyncDriver()
        mem = MemoryDriver()
        done = False

        async def task() -> None:
            nonlocal done
            await Async(tad).repeatedly(Scheduler(mem), EverySecond(1), step)
            done = True

        tad.runAsync(task())
        for ignored in range(threshold):
            mem.advance()
        self.assertTrue(done)
        self.assertEqual(count, threshold)

    def test_cancel(self) -> None:
        tad = TwistedAsyncDriver()
        mem = MemoryDriver()
        succeeding: int = 0
        repeatCall: Deferred[None] | None = None
        pending: Deferred[None]

        def canceled(d: Deferred[None]) -> None:
            return

        pending = Deferred(canceled)

        async def bonk(d: Deferred[None]) -> None:
            # odd idiom for suppressing cancellation to work around
            # https://github.com/nedbat/coveragepy/issues/1595#issuecomment-1931494916
            await d.addErrback(lambda e: e.trap(CancelledError))

        async def asynchronously() -> None:
            nonlocal succeeding
            if succeeding:
                succeeding = succeeding - 1
                await pending
            else:
                await bonk(pending)

        async def synchronously() -> None:
            pass

        def go(how: Callable[[], Any]) -> None:
            nonlocal repeatCall
            repeatCall = Async(tad).repeatedly(
                Scheduler(mem),
                EverySecond(1),
                lambda times, stopper: how(),
            )

        async def run(how: Callable[[], Any]) -> None:
            go(how)
            assert repeatCall is not None, "repeatCall should already be set"
            await bonk(repeatCall)

        tad.runAsync(run(asynchronously))
        self.assertTrue(mem.isScheduled())
        assert repeatCall is not None

        repeatCall.cancel()
        self.assertFalse(mem.isScheduled())
        pending = Deferred(canceled)
        succeeding += 1
        tad.runAsync(run(asynchronously))
        self.assertTrue(mem.isScheduled())
        mem.advance()
        self.assertFalse(mem.isScheduled())
        p, pending = pending, Deferred(canceled)
        p.callback(None)
        self.assertTrue(mem.isScheduled())
        mem.advance()
        repeatCall.cancel()
        tad.runAsync(run(synchronously))
        self.assertTrue(mem.isScheduled())
        repeatCall.cancel()
        self.assertFalse(mem.isScheduled())

    def test_eachWeekOn(self) -> None:
        """
        L{EachWeekOn} provides a recurrence on custom weekdays at custom
        times.
        """
        tad = TwistedAsyncDriver()
        mem = MemoryDriver()
        mem.advance(1706826915.372823)

        TZ = ZoneInfo("America/Los_Angeles")
        dtd = DateTimeDriver(mem, TZ)
        sch = Scheduler[DateTime[ZoneInfo], Callable[[], None]](dtd)
        x = []

        async def record(
            steps: list[DateTime[ZoneInfo]], stopper: Cancellable
        ) -> None:
            x.append((sch.now(), steps, stopper))

        rule: RecurrenceRule[DateTime[ZoneInfo], list[DateTime[ZoneInfo]]] = (
            EachWeekOn(
                days={Day.MONDAY, Day.WEDNESDAY, Day.FRIDAY},
                hour=15,
                minute=10,
            )
        )

        Async(tad).repeatedly(sch, rule, record)

        mem.advance()
        mem.advance()
        mem.advance(86401 * 5)
        mem.advance(86401 * 24)

        [
            (start, _, _),
            (first, tries1, _),
            (second, tries2, _),
            (fourth, tries4, _),
            *rest,
        ] = x
        self.assertEqual(tries1, [datetime(2024, 2, 2, 15, 10, tzinfo=TZ)])
        self.assertEqual(tries2, [datetime(2024, 2, 5, 15, 10, tzinfo=TZ)])
        self.assertEqual(
            tries4,
            [
                datetime(2024, 2, 7, 15, 10, tzinfo=TZ),
                datetime(2024, 2, 9, 15, 10, tzinfo=TZ),
            ],
        )
        self.assertEqual(datetime(2024, 2, 2, 15, 10, tzinfo=TZ), first)
        self.assertEqual(datetime(2024, 2, 5, 15, 10, tzinfo=TZ), second)
        # 2/7, 2/9 skipped!
        self.assertEqual(datetime(2024, 2, 10, 15, 10, 5, tzinfo=TZ), fourth)
        bigSkip = [
            *[
                datetime(2024, 2, n, 15, 10, tzinfo=TZ)
                for n in [12, 14, 16, 19, 21, 23, 26, 28]
            ],
            *[datetime(2024, 3, n, 15, 10, tzinfo=TZ) for n in [1, 4]],
        ]
        self.maxDiff = 99999
        actual = list(chain(*[tries for (_, tries, _) in rest]))
        self.assertEqual(bigSkip, actual)



from __future__ import annotations

from asyncio.events import new_event_loop
from asyncio.exceptions import CancelledError, InvalidStateError
from contextvars import Context
from dataclasses import dataclass
from typing import Callable
from unittest import TestCase

from twisted.internet.task import Clock

from ..boundaries import Cancellable
from ..drivers.asyncio import AsyncioAsyncDriver, AsyncioTimeDriver, scheduler


@dataclass
class AsyncioClock:
    """
    Minimal translation of Twisted interface into Asyncio interface.
    """

    clock: Clock

    def call_at(
        self,
        when: float,
        callback: Callable[[], None],
        *args: object,
        context: Context | None = None,
    ) -> Cancellable:
        assert context is None, "context not yet supported"
        return self.clock.callLater(
            when - self.clock.seconds(),
            callback,
            *args,
        )

    def time(self) -> float:
        return self.clock.seconds()


class AsyncDriverTests(TestCase):
    def setUp(self) -> None:
        self.called = 0
        self.loop = new_event_loop()

    def tearDown(self) -> None:
        self.loop.close()

    def call(self) -> None:
        self.called += 1

    def test_complete(self) -> None:
        driver = AsyncioAsyncDriver(self.loop)

        f = driver.newWithCancel(self.call)
        with self.assertRaises(InvalidStateError):
            f.result()

        async def a() -> None:
            driver.complete(f)
            await f

        self.loop.run_until_complete(a())
        self.assertEqual(self.called, 0)
        self.assertIsNone(f.result())

    def test_cancel(self) -> None:
        driver = AsyncioAsyncDriver(self.loop)
        f = driver.newWithCancel(self.call)
        with self.assertRaises(InvalidStateError):
            f.result()
        self.assertEqual(self.called, 0)

        async def a() -> None:
            f.cancel()
            with self.assertRaises(CancelledError):
                await f

        self.loop.run_until_complete(a())
        self.assertEqual(self.called, 1)
        with self.assertRaises(CancelledError):
            f.result()

    def test_runAsync(self) -> None:
        driver = AsyncioAsyncDriver(self.loop)

        # Set up a bunch of futures to ensure tasks are run independently
        # because runAsync doesn't return an awaitable.

        f1 = driver.newWithCancel(self.call)
        f2 = driver.newWithCancel(self.call)
        f3 = driver.newWithCancel(self.call)
        f4 = driver.newWithCancel(self.call)
        f5 = driver.newWithCancel(self.call)
        f6 = driver.newWithCancel(self.call)
        e1 = driver.newWithCancel(self.call)
        e2 = driver.newWithCancel(self.call)
        concurrent = []

        async def c1() -> None:
            concurrent.append("c1")
            await f1
            f3.set_result(None)
            await e1
            concurrent.append("c3")
            f5.set_result(None)

        async def c2() -> None:
            concurrent.append("c2")
            await f2
            f4.set_result(None)
            await e2
            concurrent.append("c4")
            f6.set_result(None)

        async def t() -> None:
            f1.set_result(None)
            f2.set_result(None)
            await f3
            await f4
            self.assertEqual(concurrent, ["c1", "c2"])
            e1.set_result(None)
            e2.set_result(None)
            await f5
            await f6
            self.assertEqual(concurrent, ["c1", "c2", "c3", "c4"])

        driver.runAsync(c1())
        driver.runAsync(c2())

        self.loop.run_until_complete(t())
        self.assertEqual(self.called, 0)


class TimeDriverTests(TestCase):
    def setUp(self) -> None:
        self.called = 0

    def call(self) -> None:
        self.called += 1

    def test_schedule(self) -> None:
        driver = AsyncioTimeDriver(AsyncioClock(clock := Clock()))
        driver.reschedule(3.0, self.call)
        self.assertEqual(self.called, 0)
        clock.advance(4.0)
        self.assertEqual(self.called, 1)

    def test_now(self) -> None:
        driver = AsyncioTimeDriver(AsyncioClock(clock := Clock()))
        clock.advance(7.2)
        self.assertEqual(driver.now(), 7.2)

    def test_reschedule(self) -> None:
        driver = AsyncioTimeDriver(AsyncioClock(clock := Clock()))
        driver.reschedule(3.0, self.call)
        self.assertEqual(self.called, 0)
        driver.reschedule(5.0, self.call)
        clock.advance(4.0)
        self.assertEqual(self.called, 0)
        clock.advance(1.0)
        self.assertEqual(self.called, 1)

    def test_unschedule(self) -> None:
        driver = AsyncioTimeDriver(AsyncioClock(clock := Clock()))
        driver.reschedule(3.0, self.call)
        self.assertEqual(self.called, 0)
        driver.unschedule()
        clock.advance(4.0)
        self.assertEqual(self.called, 0)
        driver.unschedule()  # safe no-op

    def test_scheduler(self) -> None:
        sched = scheduler(AsyncioClock(clock := Clock()))
        stuff = []

        def hello() -> None:
            stuff.append("hello")

        sched.callAt(50, hello)
        self.assertEqual(stuff, [])
        clock.advance(60)
        self.assertEqual(stuff, ["hello"])

    def test_schedulerDefaults(self) -> None:
        sched = scheduler()
        self.assertIsInstance(sched.driver, AsyncioTimeDriver)



"""
Implementation of L{TimeDriver} and L{AsyncDriver} in terms of U{Twisted
<https://twisted.org/>}'s APIs, L{IReactorTime} and L{Deferred}.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Coroutine, Optional

from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IDelayedCall, IReactorTime
from twisted.logger import Logger

from ..boundaries import AsyncDriver, PriorityQueue, TimeDriver
from ..heap import Heap
from ..scheduler import FutureCall, Scheduler, SimpleScheduler

log = Logger()


@dataclass
class TwistedTimeDriver:
    """
    Instantiate a L{TwistedTimeDriver} with an L{IReactorTime}; for example::

        from twisted.internet.task import react

        async def main(reactor):
            driver = TwistedTimeDriver(reactor)
            ...

        task.react(main)
    """

    _reactor: IReactorTime
    _call: Optional[IDelayedCall] = None

    def reschedule(self, desiredTime: float, work: Callable[[], None]) -> None:
        """"""

        def _() -> None:
            self._call = None
            work()

        if self._call is not None:
            self._call.cancel()
        self._call = self._reactor.callLater(
            max(0, desiredTime - self.now()), _
        )

    def unschedule(self) -> None:
        if self._call is not None:
            self._call.cancel()
            self._call = None

    def now(self) -> float:
        return self._reactor.seconds()


_TimeDriverCheck: type[TimeDriver[float]] = TwistedTimeDriver


@dataclass
class TwistedAsyncDriver:
    """
    Driver for Deferred-flavored awaitables.
    """

    def newWithCancel(self, cancel: Callable[[], None]) -> Deferred[None]:
        """
        Create a new future-ish object with the given callback to execute when
        canceled.
        """
        return Deferred(lambda d: cancel())

    def complete(self, asyncObj: Deferred[None]) -> None:
        """
        The asynchronous operation completed successfully.
        """
        asyncObj.callback(None)

    def runAsync(
        self, coroutine: Coroutine[Deferred[None], Any, Any]
    ) -> Deferred[None]:
        return Deferred.fromCoroutine(coroutine)


_AsyncDriverCheck: type[AsyncDriver[Deferred[None]]] = TwistedAsyncDriver


def scheduler(
    reactor: IReactorTime | None = None,
    queue: PriorityQueue[FutureCall[float, Callable[[], None]]] | None = None,
) -> SimpleScheduler:
    """
    Create a scheduler that uses Twisted.
    """
    if reactor is None:
        from twisted.internet import reactor  # type:ignore[assignment]

        assert reactor is not None
    return Scheduler(
        TwistedTimeDriver(reactor),
        queue if queue is not None else Heap(),
    )



from fritter.drivers.asyncio import AsyncioTimeDriver
from fritter.scheduler import SimpleScheduler
from asyncio import run, Future


async def example() -> None:
    s = SimpleScheduler(AsyncioTimeDriver())
    f = Future[None]()

    def bye() -> None:
        f.set_result(None)

    start = s.now()
    s.callAt(start + 1.5, bye)
    await f
    end = s.now()
    print(f"elapsed={end-start}")


run(example())



# setup
from typing import Callable
from fritter.drivers.memory import MemoryDriver
from fritter.scheduler import SimpleScheduler
from fritter.tree import branch

driver = MemoryDriver()
trunk = SimpleScheduler(driver)
manager, branched = branch(trunk)
# end setup


# showfunc
def show(name: str) -> Callable[[], None]:
    def _() -> None:
        print(f"{name} trunk={trunk.now()} branch={branched.now()}")

    return _
# end showfunc


# branchcalls
branched.callAt(1.0, show("branch 1"))
branched.callAt(2.0, show("branch 2"))
branched.callAt(3.0, show("branch 3"))
# trunkcalls
trunk.callAt(1.0, show("trunk 1"))
trunk.callAt(2.0, show("trunk 2"))
trunk.callAt(3.0, show("trunk 3"))
# endcalls

# interact
driver.advance()
print("pause")
manager.pause()
driver.advance()
print("unpause")
manager.unpause()
driver.advance()
driver.advance()



"""
In-memory implementation of L{TimeDriver} for use in tests and batch scripts.
"""

from __future__ import annotations

from dataclasses import dataclass
from math import inf, nextafter
from typing import Callable, Optional, Tuple

from ..boundaries import TimeDriver


@dataclass
class MemoryDriver:
    """
    In-memory L{TimeDriver} that only moves when L{advance
    <MemoryDriver.advance>} is called.
    """

    _currentTime: float = 0.0
    _scheduledWork: Optional[Tuple[float, Callable[[], None]]] = None

    def reschedule(self, desiredTime: float, work: Callable[[], None]) -> None:
        """
        Schedule the given work to happen at the given time.

        @note: In order to avoid infinite loops where time is not moving
            forward while work is being scheduled at the current moment or in
            the past, time (as referenced by L{MemoryDriver.now}) will always
            advance by at least one U{ULP
            <https://en.wikipedia.org/wiki/Unit_in_the_last_place>}, so if
            C{desiredTime} is before or exactly C{now}, by the time C{work} is
            run, C{now} will be very slightly greater.

        @see: L{TimeDriver.reschedule}
        """
        minInterval = nextafter(self._currentTime, inf)
        self._scheduledWork = max(minInterval, desiredTime), work

    def unschedule(self) -> None:
        "L{TimeDriver.unschedule}"
        self._scheduledWork = None

    def now(self) -> float:
        "L{TimeDriver.now}"
        return self._currentTime

    # |   memory driver only  |
    # v                       v

    def advance(self, delta: Optional[float] = None) -> float | None:
        """
        Advance the clock of L{this driver <MemoryDriver>} by C{delta} seconds.

        If no C{delta} is provided, then advance until the next scheduled time
        when this driver would run something.

        @return: the amount of time that was advanced, or None if no work was
            scheduled.
        """
        if delta is None:
            if self._scheduledWork is not None:
                delta = max(0, self._scheduledWork[0] - self._currentTime)
            else:
                return None
        self._currentTime += delta
        while (self._scheduledWork is not None) and (
            self._currentTime >= self._scheduledWork[0]
        ):
            what = self._scheduledWork[1]
            self._scheduledWork = None
            what()
        return delta

    def isScheduled(self) -> bool:
        """
        Does this driver currently have work scheduled with it?
        """
        return self._scheduledWork is not None


_DriverTypeCheck: type[TimeDriver[float]] = MemoryDriver



from datetime import datetime
from zoneinfo import ZoneInfo

from datetype import aware
from fritter.drivers.datetime import DateTimeDriver, guessLocalZone
from fritter.drivers.memory import MemoryDriver
from fritter.persistent.json import JSONableScheduler, JSONRegistry

ctxtype = dict[str, str]

registry = JSONRegistry[ctxtype]()


@registry.function
def call1() -> None:
    print("hello world")


memoryDriver = MemoryDriver()
scheduler = JSONableScheduler[ctxtype](DateTimeDriver(memoryDriver))
dt = aware(
    datetime(
        2023,
        7,
        21,
        1,
        1,
        1,
        tzinfo=guessLocalZone(),
    ),
    ZoneInfo,
)
handle = scheduler.callAt(dt, call1)
dump = registry.save(scheduler)
print(dump)
mem2 = MemoryDriver()
loaded = registry.load(mem2, dump, {})
mem2.advance(dt.timestamp())



# -*- test-case-name: fritter.test.test_sleep -*-
"""
Implementation of L{TimeDriver} that can run timers by blocking until all its
timers have been run, sleeping between them as necessary.

This is suitable for batch scripts that don't require an event loop like
L{asyncio <fritter.drivers.asyncio>} or L{twisted <fritter.drivers.twisted>}.
"""
from dataclasses import dataclass
from math import inf
from time import sleep as _sleep
from time import time as _time
from typing import TYPE_CHECKING, Callable

from ..boundaries import TimeDriver


@dataclass
class SleepDriver:
    """
    Instantiate a L{SleepDriver} with no arguments to get one that sleeps using
    L{time.sleep} and gets the current time with L{time.time}.

    For testing, you can supply those parameters, but for most test cases, you
    should probably prefer a L{fritter.drivers.memory.MemoryDriver}.

    @ivar sleep: The L{time.sleep}-like callable that this driver will use to
        sleep for a given number of seconds.

    @ivar time: The L{time.time}-like callable that this driver will use to get
        the current time.
    """

    sleep: Callable[[float], None] = _sleep
    time: Callable[[], float] = _time
    _work: tuple[float, Callable[[], None]] | None = None

    def reschedule(self, desiredTime: float, work: Callable[[], None]) -> None:
        "Implementation of L{TimeDriver.reschedule}"
        self._work = desiredTime, work

    def unschedule(self) -> None:
        "Implementation of L{TimeDriver.unschedule}"
        self._work = None

    def now(self) -> float:
        "Implementation of L{TimeDriver.now}"
        return self.time()

    def block(self, timeout: float = inf) -> int:
        """
        While any active timer is scheduled with L{reschedule
        <SleepDriver.reschedule>}, sleep until the desired time specified by
        that call, call the work that was scheduled, and repeat.

        @param timeout: if specified, the I{maximum, total} number of seconds
            to sleep before unblocking, even if work remains to be done.

        @note: When this driver runs out of work to do, it will return
            immediately, even if it has not yet slept for C{total} seconds.  An
            idle L{SleepDriver} will return immediately without ever blocking.
        """
        worked = 0
        maxTime = self.time() + timeout
        while True:
            scheduled = self._work
            if scheduled is None:
                break
            time, work = scheduled
            now = self.time()
            self.sleep(max(0, min(time, maxTime) - now))
            if time > maxTime:
                break
            worked += 1
            # This might get weird, or maybe even lose work, if self.sleep
            # calls self.reschedule.  But it should never do that (it's a
            # potential minefield of reentrancy weirdness, sort of but not
            # entirely like a thread-safety bug), and we might want to enforce
            # that with a runtime check.
            self._work = None
            work()
        return worked


if TYPE_CHECKING:
    _CheckSleepDriver: type[TimeDriver[float]] = SleepDriver



# -*- test-case-name: fritter.test.test_asyncio -*-

"""
Implementation of L{TimeDriver} and L{AsyncDriver} for L{asyncio}.
"""

from __future__ import annotations

from asyncio import Future, get_event_loop
from asyncio.events import AbstractEventLoop
from contextvars import Context
from dataclasses import dataclass, field
from typing import Any, Callable, Coroutine, Protocol

from ..boundaries import AsyncDriver, Cancellable, PriorityQueue, TimeDriver
from ..heap import Heap
from ..scheduler import FutureCall, Scheduler, SimpleScheduler


class LoopTimeInterface(Protocol):
    """
    Describe the portions of C{AbstractEventLoop} used by L{AsyncioTimeDriver}.
    """

    # TODO: fix code link

    def call_at(
        self,
        when: float,
        callback: Callable[[], None],
        *args: object,
        context: Context | None = None,
    ) -> Cancellable: ...

    def time(self) -> float: ...


@dataclass
class AsyncioTimeDriver:
    """
    An implementation of L{TimeDriver} using an L{asyncio} event loop.
    """

    _loop: LoopTimeInterface = field(default_factory=get_event_loop)
    _call: Cancellable | None = None

    def reschedule(self, desiredTime: float, work: Callable[[], None]) -> None:
        "Implementation of L{TimeDriver.reschedule}"

        def _() -> None:
            self._call = None
            work()

        if self._call is not None:
            self._call.cancel()
        self._call = self._loop.call_at(max(0, desiredTime), _)

    def unschedule(self) -> None:
        "Implementation of L{TimeDriver.unschedule}"
        if self._call is not None:
            self._call.cancel()
            self._call = None

    def now(self) -> float:
        "Implementation of L{TimeDriver.now}"
        return self._loop.time()


_TimeDriverCheck: type[TimeDriver[float]] = AsyncioTimeDriver


@dataclass
class AsyncioAsyncDriver:
    """
    Driver for L{Future}-flavored awaitables.

    @see: L{fritter.repeat.Async}

    @see: L{TimeDriver}
    """

    _loop: AbstractEventLoop = field(default_factory=get_event_loop)

    def newWithCancel(self, cancel: Callable[[], None]) -> Future[None]:
        """
        Create a new L{Future} with the given callback to execute when
        canceled.
        """
        f = Future[None](loop=self._loop)

        @f.add_done_callback
        def done(future: Future[None]) -> None:
            if f.cancelled():
                cancel()

        return f

    def complete(self, asyncObj: Future[None]) -> None:
        "The asynchronous operation completed successfully."
        asyncObj.set_result(None)

    def runAsync(
        self, coroutine: Coroutine[Future[None], Any, Any]
    ) -> Future[None]:
        """
        Run the given task on the event loop with L{asyncio.create_task}.
        """
        return self._loop.create_task(coroutine)


_AsyncioDriverCheck: type[AsyncDriver[Future[None]]] = AsyncioAsyncDriver


def scheduler(
    loop: LoopTimeInterface | None = None,
    queue: PriorityQueue[FutureCall[float, Callable[[], None]]] | None = None,
) -> SimpleScheduler:
    """
    Create a scheduler that uses Asyncio.
    """
    return Scheduler(
        AsyncioTimeDriver(loop if loop is not None else get_event_loop()),
        queue if queue is not None else Heap(),
    )
